/*
 * S I L E N C E
 *
 * An auto-extensible system for making music on computers by means of
 * software alone.
 * Copyright (c) 2001 by Michael Gogins. All rights reserved.
 *
 * L I C E N S E
 *
 * This software is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * P U R P O S E
 *
 * Defines the public C application programming interface to Csound.
 *
 * 29 May 2002 - ma++ merge with CsoundLib.
 * 30 May 2002 - mkg add csound "this" pointer argument back into merge.
 * 27 Jun 2002 - mkg complete Linux dl code and Makefile
 * 30 Jan 2003 - IV rewrote most of this file
 * 08 Feb 2003 - IV added many new API functions and fixed some existing ones
 */
#ifdef __cplusplus
extern "C" {
#endif

#define CS_H_NO_MACROS_FOR_GLOBALS 1

#include <cs.h>
#include "timers.h"
#ifdef LINUX
#include <dlfcn.h>
#elif defined(WIN32)
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif

#ifdef __CYGWIN__
#undef __unix
#endif

#ifndef MAX_PATH
static const int MAX_PATH = 255;
#endif

/*
 * Csound symbols referenced in this file.
 */

#define csoundMaxExits 64

static  void    *csoundExitFuncs_[csoundMaxExits];
static  long    csoundNumExits_ = -1;
static  jmp_buf csoundJump_;

extern  int     csoundMain(void *csound, int argc, char **argv);
extern  void    csoundMainCleanup(void);
extern  void    mainRESET(void);
extern  void    dispkill(WINDAT *windat);
extern  long    kperf(long kcnt);
extern  long    kcnt;
extern  char    *inbuf;
extern  void    *outbuf;
extern  void    SfReset(void);
extern  void    csoundDefaultMidiOpen(void *csound);
extern  void    create_opcodlst(void);
extern  int     writeLine(const char *text, long size);
extern  void    newevent(/* void *csound, */ char type, MYFLT *pfields,
                         long numFields);
extern  opcodelist  *new_opcode_list(void);
extern  int     dispose_opcode_list(opcodelist *list);
extern  void    csoundDefaultExternalMidiCloseCallback(void *csound);

static  int     instance_cnt = 0;

/* for measuring elapsed time */

static RTCLOCK clk_bas = { 0UL, 0UL, 0.0, 0UL, 0UL, 0.0 };

/* This needs a stub definition in this file. */

char *getDB(void)
{
    return NULL;
}

void *csoundCreate(void *hostData)
{
    /* Currently, there is only one instance of Csound per process. */
    /* In the future, this will be dynamically allocated, and will  */
    /* need to be passed to all internals Csound functions that     */
    /* reference it. */
    if (instance_cnt) {
      csoundMessage((void*) &cglob,
                    "This CsoundAPI implementation does not (yet) "
                    "support multiple instances\n");
      return NULL;
    }
    instance_cnt++;
    globals_init((void*) &cglob);
    csoundReset((void*) &cglob);
    cglob.hostdata = hostData;
    timers_struct_init(&clk_bas);       /* start timer */
    return &cglob;
}

int csoundQueryInterface(const char *name, void **interface, int *version)
{
    if(strcmp(name, "CSOUND") == 0)
    {
      *interface = csoundCreate(NULL);
      *version = csoundGetVersion();
      return 0;
    }
    return 1;
}

void csoundDestroy(void *csound)
{
    if (!instance_cnt)
      csoundMessage(csound, "No Csound instance to destroy.\n");
    else {
      /* force reloading of Csound API function pointers */
      ((GLOBALS*) csound)->csMagic1 = 0;
      csoundReset(csound);
      --instance_cnt;
    }
}

int csoundGetVersion(void)
{
    /* should not there be a way to allow patchlevels ? */
    return (CSVERSION * 100 + CSSUBVER);
}

int csoundGetAPIVersion(void)
{
    return (APIVERSION * 100 + APISUBVER);
}

void *csoundGetHostData(void *csound)
{
    return (((GLOBALS*) csound)->hostdata);
}

void csoundSetHostData(void *csound, void *hostData)
{
    ((GLOBALS*) csound)->hostdata = hostData;
}

/*
 * PERFORMANCE
 */

extern  int     runincomponents;
extern  int     sensevents(void);
extern  int     cleanup(void);
extern  int     orcompact(void);

int csoundPerform(void *csound, int argc, char **argv)
{
    int returnValue;
    /* do we have a Csound instance to perform ? */
    if (!instance_cnt) {
      err_printf("error: csoundPerform was called without csoundCreate\n");
      return CSOUND_ERROR;
    }
    runincomponents = 0;
    returnValue = csoundMain(csound, argc, argv);
    if (returnValue) {
      err_printf("Early return from csoundPerform().\n");
    }
    return returnValue;
}

int csoundCompile(void *csound, int argc, char **argv)
{
    int returnValue;
    /* do we have a Csound instance to perform ? */
    if (!instance_cnt) {
      err_printf("error: csoundCompile was called without csoundCreate\n");
      return CSOUND_ERROR;
    }
    runincomponents = 1;
    returnValue = csoundMain(csound, argc, argv);
    if (returnValue) {
      err_printf("Early return from csoundCompile().\n");
    }
    return returnValue;
}

int csoundPerformKsmps(void *csound)
{
    int done = 0;
    volatile int returnValue;

    /* setup jmp for return after an exit() */
    if ((returnValue = setjmp(csoundJump_)) != 0) {
      csoundMessage(csound, "Early return from csoundPerformKsmps().\n");
      return returnValue;
    }
    /* FIXME: allow processing multiple RTevents in a single k-cycle */
    done = sensevents();
    if (!done && kcnt)
    {
      /*
        Rather than overriding real-time event handling in kperf,
        turn it off before calling kperf, and back on afterwards.
       */
      int rtEvents = ((GLOBALS*) csound)->O.RTevents;
      ((GLOBALS*) csound)->O.RTevents = 0;
      kperf(1);
      kcnt -= 1;
      ((GLOBALS*) csound)->O.RTevents = rtEvents;
    }
    return done;
}

/* external host's outbuffer passed in csoundPerformBuffer() */

#if 0
static  char    *_rtCurOutBuf       = NULL;
static  long    _rtCurOutBufSize    = 0;
static  long    _rtCurOutBufCount   = 0;
static  char    *_rtOutOverBuf      = NULL;
static  long    _rtOutOverBufSize   = 0;
static  long    _rtOutOverBufCount  = 0;
static  char    *_rtInputBuf        = NULL;
static  long    _rtInputBufSize     = 0;
static  long    _rtInputBufIndex    = 0;
#endif

int csoundPerformBuffer(void *csound)
{
    IGN(csound);
    printf("csoundPerformBuffer(): DO NOT USE THIS INTERFACE\n");
    return(-1);
#if 0
    volatile int returnValue;
    /* Number of samples still needed to create before returning. */
    static int sampsNeeded = 0;
    int sampsPerKperf = ksmps * nchnls;
    int done = 0;
    /* Setup jmp for return after an exit().
    */
    if (returnValue = setjmp(csoundJump_)) {
      csoundMessage(csound, "Early return from csoundPerformBuffer().");
      return returnValue;
    }
    _rtCurOutBufCount = 0;
    sampsNeeded += CSGLOBAL(O).outbufsamps;
    while (!done && sampsNeeded > 0) {
      done = sensevents();
      if (done) {
        return done;
      }
      if (kcnt) {
        int rtEvents = CSGLOBAL(O).RTevents;
        CSGLOBAL(O).RTevents = 0;
        kperf(1);
        kcnt -= 1;
        sampsNeeded -= sampsPerKperf;
        CSGLOBAL(O).RTevents = rtEvents;
      }
    }
    return done;
#endif
}

void csoundCleanup(void *csound)
{
    void remove_tmpfiles(void);
    IGN(csound);
    orcompact();
    cleanup();
    remove_tmpfiles();
    /* Call all the funcs registered with atexit(). */
    while (csoundNumExits_ >= 0)
    {
      void (*func)(void) = (void (*)(void)) csoundExitFuncs_[csoundNumExits_];
      func();
      csoundNumExits_--;
    }
}

/*
 * ATTRIBUTES
 */

CSGLOBALS *csoundGetGlobals(void *csound)
{
    return ((CSGLOBALS*) &(((GLOBALS*) csound)->csglobals));
}

MYFLT csoundGetSr(void *csound)                 /* sample rate */
{
    return ((GLOBALS*) csound)->csglobals.esr;
}

MYFLT csoundGetKr(void *csound)                 /* control rate of current */
{                                               /* instrument instance */
    return ((GLOBALS*) csound)->csglobals.ekr;
}

MYFLT csoundGetGlobalKr(void *csound)           /* orchestra level */
{                                               /* control rate */
    return ((GLOBALS*) csound)->csglobals.global_ekr;
}

int csoundGetKsmps(void *csound)                /* ksmps of current */
{                                               /* instrument instance */
    return (int) ((GLOBALS*) csound)->csglobals.ksmps;
}

int csoundGetGlobalKsmps(void *csound)          /* orchestra level ksmps */
{
    return (int) ((GLOBALS*) csound)->csglobals.global_ksmps;
}

int csoundGetNchnls(void *csound)               /* number of channels */
{                       /* should we assume input is same as output ? */
    return (int) ((GLOBALS*) csound)->csglobals.nchnls;
}

MYFLT csoundGet0dBFS(void *csound)              /* 0dBFS level */
{
    return ((GLOBALS*) csound)->csglobals.e0dbfs;
}

int csoundGetSampleFormat(void *csound)         /* output sample format */
{
    /* should we assume input is same as output ? */
    return (int) ((GLOBALS*) csound)->O.outformat;
}

int csoundGetSampleSize(void *csound)           /* output sample size */
{
    /* should we assume input is same as output ? */
    return (int) ((GLOBALS*) csound)->O.outsampsiz;
}

long csoundGetInputBufferSize(void *csound)     /* input buffer size */
{                                               /* multiplied by nchnls */
    return (long) ((GLOBALS*) csound)->O.inbufsamps;
}

long csoundGetOutputBufferSize(void *csound)    /* output buffer size */
{                                               /* multiplied by nchnls */
    return (long) ((GLOBALS*) csound)->O.outbufsamps;
}

void *csoundGetInputBuffer(void *csound)        /* return pointer to */
{                                               /* input buffer */
    /* WARNING: if the audio implementation uses callbacks, this pointer */
    /* may change during performance */
    return ((void*) ((GLOBALS*) csound)->inbuf);
}

void *csoundGetOutputBuffer(void *csound)       /* return pointer to */
{                                               /* output buffer */
    /* WARNING: if the audio implementation uses callbacks, this pointer */
    /* may change during performance */
    return ((void*) ((GLOBALS*) csound)->outbuf);
}

MYFLT *csoundGetSpin(void *csound)
{
    return ((GLOBALS*) csound)->csglobals.spin;
}

MYFLT *csoundGetSpout(void *csound)
{
    return ((GLOBALS*) csound)->csglobals.spout;
}

MYFLT csoundGetScoreTime(void *csound)
{
    /* is it expected to be in seconds ? The documentation is not clear ... */
    return (MYFLT) ((double) ((GLOBALS*) csound)->csglobals.global_kcounter
                    * (double) ((GLOBALS*) csound)->csglobals.global_onedkr);
}

MYFLT csoundGetProgress(void *csound)
{
    IGN(csound);
    return FL(0.0);     /* do not know */
}

/* score time / real time */

MYFLT csoundGetProfile(void *csound)
{
    MYFLT t;

    t = (MYFLT) timers_get_real_time(&clk_bas);
    if (t <= FL(0.0))
      return (FL(1.0));
    t = (MYFLT) ((double) ((GLOBALS*) csound)->csglobals.global_kcounter
                 * (double) ((GLOBALS*) csound)->csglobals.global_onedkr) / t;
    return t;
}

/* CPU time / real time */

MYFLT csoundGetCpuUsage(void *csound)
{
    MYFLT t;
    IGN(csound);
    t = (MYFLT) timers_get_real_time(&clk_bas);
    if (t <= FL(0.0))
      return (FL(1.0));
    t = (MYFLT) timers_get_CPU_time(&clk_bas) / t;
    return t;
}

/*
 * SCORE HANDLING
 */

static int csoundIsScorePending_ = 0;

int csoundIsScorePending(void *csound)
{
    IGN(csound);
    return csoundIsScorePending_;
}

void csoundSetScorePending(void *csound, int pending)
{
    IGN(csound);
    csoundIsScorePending_ = pending;
}

static MYFLT csoundScoreOffsetSeconds_ = (MYFLT) 0.0;

void csoundSetScoreOffsetSeconds(void *csound, MYFLT offset)
{
    IGN(csound);
    csoundScoreOffsetSeconds_ = offset;
}

MYFLT csoundGetScoreOffsetSeconds(void *csound)
{
    IGN(csound);
    return csoundScoreOffsetSeconds_;
}

void csoundRewindScore(void *csound)
{
    IGN(csound);
    if (((GLOBALS*) csound)->scfp) {
      fseek(((GLOBALS*) csound)->scfp, 0, SEEK_SET);
    }
}

static void csoundDefaultMessageVCallback(void *csound, const char *format,
                                          va_list args)
{
    IGN(csound);
    vfprintf(stderr, format, args);
}

static void csoundDefaultMessageSCallback(void *csound, const char *format,
                                          va_list args)
{
    IGN(csound);
    vfprintf(stdout, format, args);
}

static void (*csoundMessageVCallback_)(void *csound, const char *format,
                                       va_list args) =
    csoundDefaultMessageVCallback;

static void (*csoundMessageSCallback_)(void *csound, const char *format,
                                       va_list args) =
    csoundDefaultMessageSCallback;

/* set both message callbacks */

void csoundSetMessageCallback(void *csound,
                              void (*csoundMessageCallback)
                                   (void *csound, const char *format,
                                    va_list args))
{
    IGN(csound);
    csoundMessageVCallback_ = csoundMessageCallback;
    csoundMessageSCallback_ = csoundMessageCallback;
}

/* set stderr callback only */

void csoundSetMessageVCallback(void *csound,
                               void (*csoundMessageCallback)
                                    (void *csound, const char *format,
                                     va_list args))
{
    IGN(csound);
    csoundMessageVCallback_ = csoundMessageCallback;
}

/* set stdout callback only */

void csoundSetMessageSCallback(void *csound,
                               void (*csoundMessageCallback)
                                    (void *csound, const char *format,
                                     va_list args))
{
    IGN(csound);
    csoundMessageSCallback_ = csoundMessageCallback;
}

void csoundMessageV(void *csound, const char *format, va_list args)
{
    csoundMessageVCallback_(csound, format, args);
}

void csoundMessageS(void *csound, const char *format, va_list args)
{
    csoundMessageSCallback_(csound, format, args);
}

void csoundMessage(void *csound, const char *format, ...)
{
    va_list args;
    va_start(args, format);
    csoundMessageSCallback_(csound, format, args);
    va_end(args);
}

/* for compatibility */

void dribble_printf(const char *format, ...)
{
    va_list args;
    if (cglob.dribble != NULL) {
      va_start(args, format);
      vfprintf(cglob.dribble, format, args);
      va_end(args);
    }
    va_start(args, format);
    csoundMessageSCallback_(&cglob, format, args);
    va_end(args);
}

void err_printf(const char *format, ...)
{
    va_list args;
    va_start(args, format);
    csoundMessageVCallback_(&cglob, format, args);
    va_end(args);
    if (cglob.dribble != NULL) {
      va_start(args, format);
      vfprintf(cglob.dribble, format, args);
      va_end(args);
    }
}

/* not sure what these do */

static void (*csoundThrowMessageCallback_)(void *csound, const char *format,
                                           va_list args) =
    csoundDefaultMessageVCallback;

void csoundSetThrowMessageCallback(void *csound,
                                   void (*csoundThrowMessageCallback)
                                        (void *csound, const char *format,
                                         va_list args))
{
    IGN(csound);
    csoundThrowMessageCallback_ = csoundThrowMessageCallback;
}

void csoundThrowMessageV(void *csound, const char *format, va_list args)
{
    csoundThrowMessageCallback_(csound, format, args);
}

void csoundThrowMessage(void *csound, const char *format, ...)
{
    va_list args;
    va_start(args, format);
    csoundThrowMessageCallback_(csound, format, args);
    va_end(args);
}

/* get / set message level */

int csoundGetMessageLevel(void *csound)
{
    return ((GLOBALS*) csound)->O.msglevel;
}

void csoundSetMessageLevel(void *csound, int messageLevel)
{
    ((GLOBALS*) csound)->O.msglevel = messageLevel;
}

void csoundInputMessage(void *csound, const char *message)
{
    IGN(csound);
    writeLine(message, strlen(message));
}

  /*
   * sensekey implementation
   */

#define INCHARBUF_SIZE  256
#define INCHARBUF_MASK  255

static char inChar_buf[INCHARBUF_SIZE];
static int  inChar_buf_wrptr = 0;
static int  inChar_buf_rdptr = 0;

#ifdef WIN32
#include <conio.h>
#elif defined(__unix) || defined(LINUX)
#include <termios.h>
#include <unistd.h>
#include <fcntl.h>

static  struct termios  tty;
static  tcflag_t        saved_c_lflag;
static  int             tty_attr_set = 0;   /* are terminal attributes set ? */

static void restore_tty_flags(void)
{
    if (!tty_attr_set)
      return;           /* nothing to restore */
    tty_attr_set = 0;
    tcgetattr(0, &tty);
    tty.c_lflag = saved_c_lflag;
    tcsetattr(0, TCSANOW, &tty);
}

static void inchar_tty_init(void)
{
    if (!tty_attr_set) {
      long flgs;
      tcgetattr(0, &tty);
      saved_c_lflag = tty.c_lflag;
      tty.c_lflag &= (~(ICANON | ECHO));
      tcsetattr(0, TCSANOW, &tty);
      atexit(restore_tty_flags);        /* restore terminal on exit */
      /* switch to non-blocking mode */
      flgs = (long) fcntl(0, F_GETFL);
      fcntl(0, F_SETFL, flgs | (long) O_NONBLOCK);
      tty_attr_set = 1;
    }
}
#endif

void csoundKeyPress(void *csound, char c)
{
    IGN(csound);
    /* update buffer write pointer and wrap around if necessary */
    inChar_buf_wrptr = ((inChar_buf_wrptr + 1) & INCHARBUF_MASK);
    if (inChar_buf_wrptr == inChar_buf_rdptr) {
      err_printf("WARNING: buffer overrun in csoundKeyPress()\n");
      /* drop oldest character */
      inChar_buf_rdptr = ((inChar_buf_rdptr + 1) & INCHARBUF_MASK);
    }
    /* write new character to circular buffer */
    inChar_buf[inChar_buf_wrptr] = c;
}

char csoundGetChar(void)
{
    char  c = '\0';
    /* check if there are any characters in the buffer */
    if (inChar_buf_wrptr != inChar_buf_rdptr) {
      /* if yes, return oldest character */
      c = inChar_buf[inChar_buf_rdptr];
      /* update buffer read pointer */
      inChar_buf_rdptr = ((inChar_buf_rdptr + 1) & INCHARBUF_MASK);
    }
    else {
#ifdef WIN32
      if (_kbhit()) {
        c = (char) ((int) _getch() /* & 0x7F */);
        if (c == '\r')
          c = '\n';             /* convert CR */
      }
#elif defined(__unix) || defined(LINUX)
      if (!tty_attr_set)
        inchar_tty_init();          /* initialise terminal if not done yet */
      /* otherwise poll stdin for input (currently UNIX only) */
      unsigned char ch;
      /* IV - Jan 30 2003 */
      if ((int) read(0, &ch, 1) > 0 && (int) ch != '\0')
        c = (char) ch;
#endif
    }
    /* return whatever was read */
    return c;
}

  /*
   * CONTROL AND EVENTS
   */

static void (*csoundInputValueCallback_)(void *csound, char *channelName,
                                         MYFLT *value) = NULL;

void csoundSetInputValueCallback(void *csound,
                                 void (*inputValueCalback)
                                      (void *csound, char *channelName,
                                       MYFLT *value))
{
    IGN(csound);
    csoundInputValueCallback_ = inputValueCalback;
}

void csoundInputValue(char *channelName, MYFLT *value)
{
    if (csoundInputValueCallback_) {
      csoundInputValueCallback_(&cglob, channelName, value);
    }
    else {
      *value = FL(0.0);
    }
}

static void (*csoundOutputValueCallback_)(void *csound, char *channelName,
                                          MYFLT value) = NULL;

void csoundSetOutputValueCallback(void *csound,
                                  void (*outputValueCalback)
                                       (void *csound, char *channelName,
                                        MYFLT value))
{
    IGN(csound);
    csoundOutputValueCallback_ = outputValueCalback;
}

void csoundOutputValue(char *channelName, MYFLT value)
{
    if (csoundOutputValueCallback_) {
      csoundOutputValueCallback_(&cglob, channelName, value);
    }
}

void csoundScoreEvent(void *csound, char type, MYFLT *pfields, long numFields)
{
    IGN(csound);
    newevent(type, pfields, numFields);
}

  /*
   * REAL-TIME AUDIO
   */

/* IV - Feb 22 2003: dummy real-time audio routines */
/* can be overridden by plugin libraries            */

static  void    *playopen_(void *csound, AUDIOPARMS *devParms)
{
    IGN(csound);
    IGN(devParms);
    return NULL;
}

static  void    rtplay_(void *audioOutData, char *outBuf, int nbytes)
{
    IGN(audioOutData);
    IGN(outBuf);
    IGN(nbytes);
}

static  void    *recopen_(void *csound, AUDIOPARMS *devParms)
{
    IGN(csound);
    IGN(devParms);
    return NULL;
}

static  int     rtrecord_(void *audioInData, char *inBuf, int nbytes)
{
    IGN(audioInData);
    memset(inBuf, 0, (size_t) nbytes);
    return nbytes;
}

static  void    rtclose_(void *audioInData, void *audioOutData)
{
    IGN(audioInData);
    IGN(audioOutData);
}

static  void    *(*playopenCallback)
                 (void *csound, AUDIOPARMS *devParms) = playopen_;
void            (*rtplay)
                 (void *audioOutData, char *outBuf, int nbytes) = rtplay_;
static  void    *(*recopenCallback)
                 (void *csound, AUDIOPARMS *devParms) = recopen_;
int             (*rtrecord)
                 (void *audioInData, char *inBuf, int nbytes) = rtrecord_;
void            (*rtclose)
                 (void *audioInData, void *audioOutData) = rtclose_;

/* playopen, recopen: wrapper functions that set up the AUDIOPARMS  */
/* structure before calling the actual audio open callbacks.        */
/* These wrappers can be removed after changing Csound sources to   */
/* use the AUDIOPARMS structures instead of various variables       */
/* scattered in GLOBALS and OPARMS.                                 */

static const int sample_sizes[] = { 1, 1, 1, 2, 4, 4, 1, 3, 8 };

void *playopen(void *csound)
{
    register GLOBALS* h = (GLOBALS*) csound;

    h->rtout_params.deviceNumber    = (int) h->rtout_dev;
    h->rtout_params.nChannels       = (int) h->csglobals.nchnls;
    h->rtout_params.sampleRate      = (float) h->csglobals.esr;
    h->rtout_params.__dummy_01      = 0.0f;
    h->rtout_params.sampleFormat    = (int) h->O.outformat;
    h->rtout_params.sampleSize      =
      (h->rtout_params.sampleFormat >= AE_CHAR &&
       h->rtout_params.sampleFormat <= AE_DOUBLE ?
       sample_sizes[h->rtout_params.sampleFormat - AE_CHAR] : 1);
    h->rtout_params.swBufSize       = (int) h->O.outbufsamps
                                      / (int) h->csglobals.nchnls;
    h->rtout_params.hwBufSize       = (int) h->O.oMaxLag;
    h->rtout_params.deviceName      = (char*) h->O.pcm_output_device;
    h->rtout_params.bufPtr          = (void*) NULL;     /* not implemented */
    h->rtout_params.schedPriority   = (int) h->O.sched_priority_level;
    h->rtout_params.sleepTime       = (int) h->O.rt_sleep_time;
    h->rtout_params.userParam1      = (int) h->O.rt_param_1;
    h->rtout_params.userParam2      = (int) h->O.rt_param_2;

    return (playopenCallback(csound, &(h->rtout_params)));
}

void *recopen(void *csound)
{
    register GLOBALS* h = (GLOBALS*) csound;

    h->rtin_params.deviceNumber     = (int) h->rtin_dev;
    h->rtin_params.nChannels        = (int) h->csglobals.nchnls;
    h->rtin_params.sampleRate       = (float) h->csglobals.esr;
    h->rtin_params.__dummy_01       = 0.0f;
    h->rtin_params.sampleFormat     = (int) h->O.informat;
    h->rtin_params.sampleSize       =
      (h->rtin_params.sampleFormat >= AE_CHAR &&
       h->rtin_params.sampleFormat <= AE_DOUBLE ?
       sample_sizes[h->rtin_params.sampleFormat - AE_CHAR] : 1);
    h->rtin_params.swBufSize        = (int) h->O.inbufsamps
                                      / (int) h->csglobals.nchnls;
    h->rtin_params.hwBufSize        = (int) h->O.oMaxLag;
    h->rtin_params.deviceName       = (char*) h->O.pcm_input_device;
    h->rtin_params.bufPtr           = (void*) NULL;     /* not implemented */
    h->rtin_params.schedPriority    = (int) h->O.sched_priority_level;
    h->rtin_params.sleepTime        = (int) h->O.rt_sleep_time;
    h->rtin_params.userParam1       = (int) h->O.rt_param_1;
    h->rtin_params.userParam2       = (int) h->O.rt_param_2;

    return (recopenCallback(csound, &(h->rtin_params)));
}

void csoundSetPlayopenCallback(void *csound,
                               void *(*playopen__)
                                     (void *csound, AUDIOPARMS *devParms))
{
    IGN(csound);
    playopenCallback = playopen__;
}

void csoundSetRtplayCallback(void *csound,
                             void (*rtplay__)
                                  (void *audioOutData,
                                   char *outBuf, int nbytes))
{
    IGN(csound);
    rtplay = rtplay__;
}

void csoundSetRecopenCallback(void *csound,
                              void *(*recopen__)
                                    (void *csound, AUDIOPARMS *devParms))
{
    IGN(csound);
    recopenCallback = recopen__;
}

void csoundSetRtrecordCallback(void *csound,
                               int (*rtrecord__)
                                   (void *audioInData,
                                    char *inBuf, int nbytes))
{
    IGN(csound);
    rtrecord = rtrecord__;
}

void csoundSetRtcloseCallback(void *csound,
                              void (*rtclose__)
                                   (void *audioInData, void *audioOutData))
{
    IGN(csound);
    rtclose = rtclose__;
}

  /*
   * MIDI -- this code replaces the MIDIDevice.c file
   */

static void (*csoundExternalMidiOpenCallback_)(void *csound) =
    csoundDefaultMidiOpen;

void csoundSetExternalMidiOpenCallback(void *csound,
                                       void (*csoundMidiOpen)(void *csound))
{
    IGN(csound);
    csoundExternalMidiOpenCallback_ = csoundMidiOpen;
}

void csoundExternalMidiOpen(void *csound)
{
    csoundExternalMidiOpenCallback_(csound);
}

static int defaultCsoundMidiRead(void *csound, unsigned char *midiData,
                                 int size)
{
    IGN(csound);
    IGN(midiData);
    IGN(size);
    return 0;
}

static int (*csoundExternalMidiReadCallback_)(void *csound,
                                              unsigned char *midiData,
                                              int size) =
    defaultCsoundMidiRead;

void csoundSetExternalMidiReadCallback(void *csound,
                                       int (*midiReadCallback)
                                           (void *csound,
                                            unsigned char *midiData, int size))
{
    IGN(csound);
    csoundExternalMidiReadCallback_ = midiReadCallback;
}

int csoundExternalMidiRead(void *csound, unsigned char *mbuf, int size)
{
    return csoundExternalMidiReadCallback_(csound, mbuf, size);
}

static int csoundExternalMidiEnabled_ = 0;

int csoundIsExternalMidiEnabled(void *csound)
{
    IGN(csound);
    return csoundExternalMidiEnabled_;
}

void csoundSetExternalMidiEnabled(void *csound, int enabled)
{
    IGN(csound);
    csoundExternalMidiEnabled_ = enabled;
}

static int defaultCsoundMidiWrite(void *csound, unsigned char *midiData)
{
    IGN(csound);
    IGN(midiData);
    return 0;
}

static int (*csoundExternalMidiWriteCallback_)(void *csound,
                                               unsigned char *midiData) =
    defaultCsoundMidiWrite;

void csoundSetExternalMidiWriteCallback(void *csound,
                                        int (*midiWriteCallback)
                                            (void *csound,
                                             unsigned char *midiData))
{
    IGN(csound);
    csoundExternalMidiWriteCallback_ = midiWriteCallback;
}

int csoundExternalMidiWrite(unsigned char *midiData)
{
    return csoundExternalMidiWriteCallback_(&cglob, midiData);
}

void csoundDefaultMidiCloseCallback(void *csound)
{
    IGN(csound);
}

static void (*csoundExternalMidiCloseCallback_)(void *csound) =
    csoundDefaultMidiCloseCallback;

void csoundSetExternalMidiCloseCallback(void *csound,
                                        void (*csoundExternalMidiCloseCallback)
                                             (void *csound))
{
    IGN(csound);
    csoundExternalMidiCloseCallback_ = csoundExternalMidiCloseCallback;
}

void csoundExternalMidiClose(void *csound)
{
    csoundExternalMidiCloseCallback_(csound);
}

  /*
   * FUNCTION TABLE DISPLAY.
   */

#ifdef WINDOWS
extern  int     Graphable_(void);
extern  void    MakeGraph_(WINDAT *, char *);
extern  void    DrawGraph_(WINDAT *);
extern  void    KillGraph_(WINDAT *);
extern  int     ExitGraph_(void);
extern  void    MakeXYin_(XYINDAT *, MYFLT, MYFLT);
extern  void    ReadXYin_(XYINDAT *);
#endif
extern  void    MakeAscii(WINDAT *, char *);
extern  void    DrawAscii(WINDAT *);
extern  void    KillAscii(WINDAT *);

static int isGraphable_ = 1;

void csoundSetIsGraphable(void *csound, int isGraphable)
{
    IGN(csound);
    isGraphable_ = isGraphable;
}

int Graphable(void)
{
#ifdef WINDOWS
    return Graphable_();
#else
    return isGraphable_;
#endif
}

static void defaultCsoundMakeGraph(void *csound, WINDAT *windat, char *name)
{
    IGN(csound);
#ifdef WINDOWS
    MakeGraph_(windat, name);
#else
    MakeAscii(windat, name);
#endif
}

static void (*csoundMakeGraphCallback_)(void *csound,  WINDAT *windat,
                                        char *name) =
    defaultCsoundMakeGraph;

void csoundSetMakeGraphCallback(void *csound,
                                void (*makeGraphCallback)
                                     (void *csound, WINDAT *windat,
                                      char *name))
{
    IGN(csound);
    csoundMakeGraphCallback_ = makeGraphCallback;
}

void MakeGraph(WINDAT *windat, char *name)
{
    csoundMakeGraphCallback_(&cglob, windat, name);
}

static void defaultCsoundDrawGraph(void *csound, WINDAT *windat)
{
    IGN(csound);
#ifdef WINDOWS
    DrawGraph_(windat);
#else
    DrawAscii(windat);
#endif
}

static void (*csoundDrawGraphCallback_)(void *csound,  WINDAT *windat) =
    defaultCsoundDrawGraph;

void csoundSetDrawGraphCallback(void *csound,
                                void (*drawGraphCallback)(void *csound,
                                                          WINDAT *windat))
{
    IGN(csound);
    csoundDrawGraphCallback_ = drawGraphCallback;
}

void DrawGraph(WINDAT *windat)
{
    csoundDrawGraphCallback_(&cglob, windat);
}

static void defaultCsoundKillGraph(void *csound, WINDAT *windat)
{
    IGN(csound);
#ifdef WINDOWS
    KillGraph_(windat);
#else
    KillAscii(windat);
#endif
}

static void (*csoundKillGraphCallback_)(void *csound, WINDAT *windat) =
    defaultCsoundKillGraph;

void csoundSetKillGraphCallback(void *csound,
                                void (*killGraphCallback)(void *csound,
                                                          WINDAT *windat))
{
    IGN(csound);
    csoundKillGraphCallback_ = killGraphCallback;
}

void KillGraph(WINDAT *windat)
{
    csoundKillGraphCallback_(&cglob, windat);
}

static int defaultCsoundExitGraph(void *csound)
{
    IGN(csound);
    return CSOUND_SUCCESS;
}

static int (*csoundExitGraphCallback_)(void *csound) = defaultCsoundExitGraph;

void csoundSetExitGraphCallback(void *csound,
                                int (*exitGraphCallback)(void *csound))
{
    IGN(csound);
    csoundExitGraphCallback_ = exitGraphCallback;
}

int ExitGraph(void)
{
    return csoundExitGraphCallback_(&cglob);
}

void MakeXYin(XYINDAT *xyindat, MYFLT x, MYFLT y)
{
#ifdef WINDOWS
    MakeXYin_(xyindat, x, y);
#else
    printf("xyin not supported. use invalue opcode instead.\n");
#endif
}

void ReadXYin(XYINDAT *xyindat)
{
#ifdef WINDOWS
    ReadXYin_(xyindat);
#else
    printf("xyin not supported. use invlaue opcodes instead.\n");
#endif
}

  /*
   * OPCODES
   */

opcodelist *csoundNewOpcodeList()
{
    create_opcodlst();
    return new_opcode_list();
}

void csoundDisposeOpcodeList(opcodelist *list)
{
    dispose_opcode_list(list);
}

int csoundAppendOpcode(void *csound,
                       char *opname,
                       int  dsblksiz,
                       int  thread,
                       char *outypes,
                       char *intypes,
                       SUBR iopadr,
                       SUBR kopadr,
                       SUBR aopadr,
                       SUBR dopadr)
{
    OENTRY **opcodlstp = &(((GLOBALS*) csound)->opcodlst);
    OENTRY **oplstendp = &(((GLOBALS*) csound)->oplstend);
    int oldSize = (int) ((uintptr_t) *oplstendp - (uintptr_t) *opcodlstp);
    int newSize = oldSize + sizeof(OENTRY);
    int oldCount = oldSize / sizeof(OENTRY);
    int newCount = oldCount + 1;
    OENTRY *oldOpcodlst = *opcodlstp;
    *opcodlstp = (OENTRY *) mrealloc(*opcodlstp, newSize);
    if (!*opcodlstp) {  /* should not happen in fact */
      *opcodlstp = oldOpcodlst;
      err_printf("Failed to allocate new opcode entry.");
      return CSOUND_ERROR;
    }
    else {
      OENTRY *oentry = *opcodlstp + oldCount;
      *oplstendp = *opcodlstp + newCount;
      oentry->opname    = opname;
      oentry->dsblksiz  = dsblksiz;
      oentry->thread    = thread;
      oentry->outypes   = outypes;
      oentry->intypes   = intypes;
      oentry->iopadr    = iopadr;
      oentry->kopadr    = kopadr;
      oentry->aopadr    = aopadr;
      oentry->dopadr    = dopadr;
      if (((GLOBALS*) csound)->O.msglevel & 0x100000) {
        printf(
          Str("Appended opcodlst[%d]: opcode = %s intypes = %s outypes = %s\n"),
              oldCount,
              oentry->opname,
              oentry->intypes,
              oentry->outypes);
      }
      return CSOUND_SUCCESS;
    }
}

int csoundOpcodeCompare(const void *v1, const void *v2)
{
    return strcmp(((OENTRY*)v1)->opname, ((OENTRY*)v2)->opname);
}

void csoundOpcodeDeinitialize(void *csound, INSDS *ip_)
{
    OPDS *p;
    IGN(csound);

    p = (OPDS*) ip_->nxts;      /* IV - Feb 22 2003 */
    while (p) {
      (*(p->opds_ex->dopadr))(p);
      p = p->opds_ex->nxts;
    }
}

  /*
   * MISC FUNCTIONS
   */

  /*
   * to enable csoundYield(), you must define in cs.h:
   * #define POLL_EVENTS() csoundYield()
   */

static int defaultCsoundYield(void *csound)
{
    IGN(csound);
    return 1;
}

static int (*csoundYieldCallback_)(void *csound) = defaultCsoundYield;

void csoundSetYieldCallback(void *csound, int (*yieldCallback)(void *csound))
{
    IGN(csound);
    csoundYieldCallback_ = yieldCallback;
}

int csoundYield(void *csound)
{
    return csoundYieldCallback_(csound);
}

static const int MAX_ENVIRONS = 128;

typedef struct Environs
{
    char *name;
    char *value;
} Environs;

static Environs *csoundEnv_ = NULL;

static int csoundNumEnvs_ = 0;

void csoundSetEnv(void *csound, const char *name, const char *value)
{
    int i = 0, l;
    IGN(csound);
    if (!name || !value)
      return;

    if (csoundEnv_ == NULL) {
      csoundEnv_ = (Environs *) mcalloc(MAX_ENVIRONS * sizeof(Environs));
    }
    l = strlen(value) + 1;
    for (i = 0; i < csoundNumEnvs_; i++) {
      if (strcmp(csoundEnv_[i].name, name) == 0) {
        mrealloc(csoundEnv_[i].value, l);
        strcpy(csoundEnv_[i].value, value);
        return;
      }
    }
    if (i >= MAX_ENVIRONS) {
      ((CSOUNDAPI*) csound)->Die(csound,
        "Exceeded maximum number of environment values");
      return;
    }

    csoundNumEnvs_++;
    csoundEnv_[i].name = mmalloc(strlen(name) + 1);
    strcpy(csoundEnv_[i].name, name);
    csoundEnv_[i].value = mmalloc(l);
    strcpy(csoundEnv_[i].value, value);
}

#ifdef getenv
#undef getenv
#endif

char *csoundGetEnv(const char *name)
{
    int   i;
    char  *s;

    /* first, try the local database */
    if (csoundEnv_) {
      for (i = 0; i < csoundNumEnvs_; i++) {
        if (strcmp(csoundEnv_[i].name, name) == 0) {
          return (csoundEnv_[i].value);
        }
      }
    }
    /* not found, continue search with system environment variables */
    s = getenv(name);
    /* and return whatever was (or was not) found */
    return s;
}

void csoundReset(void *csound)
{
    IGN(csound);
    SfReset();
    mainRESET();
    csoundIsScorePending_ = 1;
    csoundScoreOffsetSeconds_ = (MYFLT) 0.0;
}

#ifdef INGALLS
  /*
  * Returns a non-zero to the host,
  * which may call csoundCleanup().
  */
void exit(int status)
{
    longjmp(csoundJump_, status +1);
}

int atexit(void (*func)(void))
{
    if (++csoundNumExits_ < csoundMaxExits)
    {
      csoundExitFuncs_[csoundNumExits_] = func;
      return 0;
    }
    else
    {
      return -1;
    }
}
#endif

  /*
   * dynamic libraries
   */

void *csoundOpenLibrary(const char *libraryPath)
{
    void  *handle = NULL;
#ifdef LINUX
    handle = (void*) dlopen (libraryPath, RTLD_NOW | RTLD_GLOBAL);
    if (handle == NULL) {
      printf("Failed to load %s: %s\n", libraryPath, dlerror());
    }
#elif defined(WIN32)
    handle = (void*) LoadLibrary(libraryPath);
    if (handle == NULL) {
      printf("Failed to load %s\n", libraryPath);
    }
#else
    IGN(libraryPath);
    printf("csoundOpenLibrary() is not supported on this platform\n");
#endif
    return handle;
}

void *csoundCloseLibrary(void *library)
{
    if (library == NULL)
      return NULL;
#ifdef LINUX
    if (dlclose((void*) library) != 0) {
      printf("could not close library %s: %s\n", library, dlerror());
    }
#elif defined(WIN32)
    if (FreeLibrary((HMODULE) library) == FALSE) {
      printf("could not close library %s\n", library);
    }
#else
    IGN(library);
    printf("csoundCloseLibrary() is not supported on this platform\n");
#endif
    return NULL;
}

void *csoundGetLibrarySymbol(void *library, const char *procedureName)
{
#ifdef LINUX
    void  *sym;
    sym = (void*) dlsym((void*) library, procedureName);
    if (dlerror() != NULL) {
      sym = NULL;
    }
    return sym;
#elif defined(WIN32)
    void *sym;
    sym = (void*) GetProcAddress((HMODULE) library, procedureName);
    return sym;
#else
    IGN(library);
    IGN(procedureName);
    printf("csoundGetLibrarySymbol() is not supported on this platform\n");
    return NULL;
#endif
}

  /*
   * error handler callbacks (initerror, perferror, die)
   */

/* init error */

extern void initerror_(void*, char*);
static void (*csoundInitErrorCallback)(void*, char*) = initerror_;

void csoundSetInitErrorCallback(void *csound,
                                void (*initErrorCallback)(void*, char*))
{
    IGN(csound);
    csoundInitErrorCallback = initErrorCallback;
}

void csoundInitError(void *csound, char *s)
{
    csoundInitErrorCallback(csound, s);
}

/* perf error */

extern void perferror_(void*, char*);
static void (*csoundPerfErrorCallback)(void*, char*) = perferror_;

void csoundSetPerfErrorCallback(void *csound,
                                void (*perfErrorCallback)(void*, char*))
{
    IGN(csound);
    csoundPerfErrorCallback = perfErrorCallback;
}

void csoundPerfError(void *csound, char *s)
{
    csoundPerfErrorCallback(csound, s);
}

/* die */

static void default_die_(void *csound, char *s)
{
    printf("%s\n", s);
    if (((GLOBALS*) csound)->dribble)
      fflush(((GLOBALS*) csound)->dribble);     /* Should not be necessary */
    longjmp(((GLOBALS*) csound)->exitjmp, 1);
}

static void (*csoundDieCallback)(void*, char*) = default_die_;

void csoundSetDieCallback(void *csound,
                          void (*dieCallback)(void*, char*))
{
    IGN(csound);
    csoundDieCallback = dieCallback;
}

void csoundDie(void *csound, char *s)
{
    csoundDieCallback(csound, s);
}

/* general error handler */

static void default_error_handler_(void *csound, int errCode, char *errMsg,
                                   void *userData)
{
    IGN(userData);
    switch (errCode) {
      case (int) CSOUND_SUCCESS:
        break;          /* What is a successful error expected to do ? */
      case (int) CSOUND_ERROR:
      case (int) CSOUND_MEMORY:
                        /* assume these are fatal */
        ((CSOUNDAPI*) csound)->Die(csound, errMsg);
        break;
      case (int) CSOUND_INITIALIZATION:
                        /* init error */
        ((CSOUNDAPI*) csound)->InitError(csound, errMsg);
        break;
      case (int) CSOUND_PERFORMANCE:
                        /* perf error */
        ((CSOUNDAPI*) csound)->PerfError(csound, errMsg);
        break;
      case CSOUND_AUXALLOC_ERROR:
                        /* opcode memory allocation (auxalloc) error */
                        /* could be the same as InitError, but this would */
                        /* require opcodes to check the pointer returned */
                        /* by auxalloc and return if it is NULL */
        ((CSOUNDAPI*) csound)->Die(csound, errMsg);
        break;
      default:          /* unknown error code */
        ((CSOUNDAPI*) csound)->Die(csound, errMsg);
    }
}

static void (*csoundErrorHandlerCallback)(void*, int, char*, void*) =
    default_error_handler_;

/* Set callback function for general error handler. */

void csoundSetErrorHandlerCallback(void *csound,
                                   void (*errorHandlerCallback)
                                        (void*, int, char*, void*))
{
    IGN(csound);
    csoundErrorHandlerCallback = errorHandlerCallback;
}

/* Call error handler. */

void csoundError(void *csound, int errCode, char *errMsg, void *userData)
{
    csoundErrorHandlerCallback(csound, errCode, errMsg, userData);
}

  /*
   * Opcode helper functions
   */

/* in all these functions, 'p' is a pointer to the data structure of */
/* the current opcode */

/* return pointer to GLOBALS structure */
/* (opaque to API applications) */

void *csoundGetEngineForOpcode(void *p)
{
    return (void*) ((OPDS*) p)->globals->csound;
}

/* return pointer to CSGLOBALS structure */
/* (contains public data) */

void *csoundGetGlobalsForOpcode(void *p)
{
    return (void*) ((OPDS*) p)->globals;
}

/* get opcode parameters and store in variables pointed by input args: */
/*   outocount: number of output arguments                             */
/*   xoutcode:  bitwise OR of                                          */
/*                1: first output argument is a-rate                   */
/*                2: second output argument is a-rate                  */
/*                4: third output argument is a-rate                   */
/*                etc.                                                 */
/*              xoutcode is set for the first 15 output arguments      */
/*   inocount:  number of input arguments                              */
/*   xincode:   bitwise OR of                                          */
/*                1: first input argument is a-rate                    */
/*                2: second input argument is a-rate                   */
/*                4: third input argument is a-rate                    */
/*                etc.                                                 */
/*              xincode is set for the first 15 input arguments        */
/*   strargs:   pointer to an array of string arguments to the opcode  */
/*              up to 4 string arguments are supported                 */
/*              an argument is a string if it has the value of the     */
/*              SSTRCOD macro                                          */

void csoundGetOpcodeParams(void *p, int *outocount, int *xoutcode,
                                    int *inocount,  int *xincode,
                                    char ***strargs)
{
    *outocount = (int) ((OPDS*) p)->optext->t.outoffs->count;
    *xoutcode  = (int) ((OPDS*) p)->optext->t.xoutcod;
    *inocount  = (int) ((OPDS*) p)->optext->t.inoffs->count;
    *xincode   = (int) ((OPDS*) p)->optext->t.xincod;
    /* need to swap lowest two bits of xincode */
    *xincode   = (*xincode & (~3)) | ((*xincode & 1) << 1)
                 | ((*xincode & 2) >> 1);
    *strargs = (char**) (int) ((OPDS*) p)->optext->t.strargs;
}

/* return pointer to MIDI channel block for instrument instance owning */
/* the calling opcode; this may be NULL in case of a score note        */

MCHNBLK *csoundGetMIDIParams(void *p)
{
    return (MCHNBLK*) ((OPDS*) p)->insdshead->m_chnbp;
}

/* query ftable *argp at init time */
/* an init error is raised in case of invalid argument */

FUNC  *csoundQueryFTableInit(void *p, MYFLT *argp)
{
    extern FUNC *ftfind(MYFLT*);
    IGN(p);
    return (ftfind(argp));
}

/* query ftable *argp at perf time */
/* a perf error is raised in case of invalid argument */

FUNC  *csoundQueryFTablePerf(void *p, MYFLT *argp)
{
    extern FUNC *ftfindp(MYFLT*);
    IGN(p);
    return (ftfindp(argp));
}

/* query deferred size ftable *argp at init time */
/* an init error is raised in case of invalid argument */

FUNC  *csoundQueryDeferredFTableInit(void *p, MYFLT *argp)
{
    extern FUNC *ftnp2find(MYFLT*);
    IGN(p);
    return (ftnp2find(argp));
}

/* get release time for currently performing instrument instance in seconds */

MYFLT csoundGetReleaseTime(void *p)
{
    return ((MYFLT) ((OPDS*) p)->insdshead->xtratim
            * (MYFLT) ((OPDS*) p)->globals->global_onedkr);
}

/* set release time (in seconds) for current instrument instance */

void csoundSetReleaseTime(void *p, MYFLT xtim)
{
    int new_xtratim;
    new_xtratim = (int)
      (xtim * ((OPDS*) p)->globals->global_ekr + FL(0.5));
    if (new_xtratim > ((OPDS*) p)->insdshead->xtratim)
      ((OPDS*) p)->insdshead->xtratim = new_xtratim;
}

/* Returns non-zero if the current instrument is releasing. */

int csoundGetReleaseFlag(void *p)
{
    return (int) ((OPDS*) p)->insdshead->relesing;
}

/* Returns non-zero if beat mode is enabled. */

int csoundGetBeatModeFlag(void *p)
{
    return (int) ((GLOBALS*) ((OPDS*) p)->globals->csound)->O.Beatmode;
}

/* Returns the turnoff time of current instrument instance in seconds or
   beats, depending on beat mode. Return value is negative if the note has
   indefinite length.
   The current time in seconds can be calculated as
     p->h.globals->global_onedkr * p->h.globals->global_kcounter
   Note: beat mode is not very well tested and may not work correctly
   in some cases.
*/

MYFLT csoundGetOffTime(void *p)
{
    if (((GLOBALS*) ((OPDS*) p)->globals->csound)->O.Beatmode)
      return (MYFLT) ((OPDS*) p)->insdshead->offbet;
    else
      return (MYFLT) ((OPDS*) p)->insdshead->offtim;
}

/* Set the turnoff time of current instrument instance in seconds or */
/* beats, depending on beat mode. */

void csoundSetOffTime(void *p, MYFLT offtim)
{
    extern void schedofftim(INSDS*);
    extern void removeofftim(INSDS*);
    INSDS *ip = ((OPDS*) p)->insdshead;

    if (((GLOBALS*) ((OPDS*) p)->globals->csound)->O.Beatmode) {
      if (ip->offbet >= FL(0.0))        /* if the note was finite length, */
        removeofftim(ip);               /* remove from offtime chain */
      ip->offbet = offtim;
      if (offtim >= FL(0.0))            /* if new length is finite, */
        schedofftim(ip);                /* re-schedule offtime */
    }
    else {
      if (ip->offtim >= FL(0.0))        /* if the note was finite length, */
        removeofftim(ip);               /* remove from offtime chain */
      ip->offtim = offtim;
      if (offtim >= FL(0.0))            /* if new length is finite, */
        schedofftim(ip);                /* re-schedule offtime */
    }
}

/* Get pointer to the array of p-fields for the current instrument  */
/* instance. Returns the number of p-fields. (Note: pflds[1] is p1) */

int csoundGetPFields(void *p, MYFLT **pflds)
{
    INSDS *ip = ((OPDS*) p)->insdshead;
    INSTRTXT **tpp = ((GLOBALS*) ((OPDS*) p)->globals->csound)->instrtxtp;
    INSTRTXT *tp = tpp[(int) ip->insno];

    *pflds = (MYFLT*) &(ip->p0);
    return (int) tp->pmax;
}

/* skip rest of init opcodes for the current instrument instance */

void csoundSkipInitOpcodes(void *p)
{
    OPDS **ids = &(((OPDS*) p)->globals->ids);
    while ((*ids)->nxti != NULL)
      (*ids) = (*ids)->nxti;
}

/* skip rest of perf opcodes for the current instrument instance */

void csoundSkipPerfOpcodes(void *p)
{
    OPDS **pds = &(((OPDS*) p)->globals->pds);
    while ((*pds)->nxtp != NULL)
      (*pds) = (*pds)->nxtp;
}

/* turn off currently performing instrument instance (with release) */
/* should be used at performance time only */

void csoundTurnOffCurrentNote(void *p)
{
    extern void xturnoff(INSDS*);
    INSDS *ip;

    ip = ((OPDS*) p)->insdshead;
    if (ip->xtratim < 1) {      /* if no release, */
      xturnoff(ip);             /* deactivate immediately, */
      csoundSkipPerfOpcodes(p); /* and skip rest of instrument */
    }
    else
      xturnoff(ip);
}

/* allocate memory for opcode (cleared to zero) */
/* should be called at i-time only */

void csoundAuxAlloc(void *p, AUXCH *auxchp, size_t nbytes)
     /* allocate an auxds, or expand an old one */
     /*    call only from init (xxxset) modules */
{
    void        *h = ((OPDS*) p)->insdshead->aux_heap;
    CSOUNDAPI   *engine = (CSOUNDAPI*) (((OPDS*) p)->insdshead->csound);
    size_t      n, m;

    n = (size_t) (nbytes + 3L) >> 2;
    if (h == NULL) {
      /* create auxalloc heap for current instrument if not done yet */
      h = ((OPDS*) p)->insdshead->aux_heap =
        engine->HeapCreate((void*) engine, MEMALLOC_DISPOSABLE);
      if (h == NULL) {
        auxchp->size = 0L; auxchp->auxp = NULL; auxchp->endp = NULL;
        return;
      }
      /* set error code for handler */
      engine->HeapSetErrorCode(h, CSOUND_AUXALLOC_ERROR);
    }
    m = (size_t) (auxchp->size + 3L) >> 2;
    auxchp->size = nbytes;                      /* store size */
    if (auxchp->auxp != NULL) {
      /* if already allocated, */
      if (n == m) goto clear_space;     /* same size: clear only */
      /* different size: reallocate */
      auxchp->auxp = (char*) engine->Realloc(h, (void*) auxchp->auxp,
                                                sizeof(int32_t) * n);
    }
    else {
      /* not allocated yet, use malloc */
      auxchp->auxp = (char*) engine->Malloc(h, sizeof(int32_t) * n);
    }
    if (auxchp->auxp == NULL) {         /* safety */
      auxchp->size = 0L; auxchp->auxp = NULL; auxchp->endp = NULL;
      return;
    }
    auxchp->endp = auxchp->auxp + nbytes;       /* store end pointer */

 clear_space:
    /* clear allocated space to zero */
    memset((int32_t*) auxchp->auxp, 0, sizeof(int32_t) * n);
}

/* open file (automatically closed when note is deactivated)          */
/* the allowed values for 'mode' are:                                 */
/*   0: open for input (read from current directory, SSDIR, or SFDIR) */
/*   1: open for output (write to current directory)                  */
/*   2: open for output (write to SFDIR)                              */
/*   3: open for output (write to SFDIR, or current directory)        */
/* returns file descriptor                                            */
/* should be called at init time only                                 */

int csoundOpenFile(void *p, FDCH *fdchp, const char *name, int mode)
{
    int fd;

    IGN(p);
    if (!mode) {
      extern int openin(char*);
      fd = openin((char*) name);
    }
    else {
      extern int openout(char*, int);
      fd = openout((char*) name, mode);
    }
    /* store file descriptor */
    if (fd > 0) {
      extern void fdrecord(FDCH*);
      fdchp->fd = fd;
      fdrecord(fdchp);
    }
    else
      fdchp->fd = 0;
    return fd;
}

/* close file opened with csoundOpenFile() */
/* should be called at init time only      */

void csoundCloseFile(void *p, FDCH *fdchp)
{
    extern void fdclose(FDCH*);
    IGN(p);
    fdclose(fdchp);
}

/* load an entire file into memory             */
/* if not fullpath, look in current directory, */
/* then SADIR (if defined).                    */

MEMFIL *csoundLoadFile(void *p, char *filnam) {
    extern MEMFIL *ldmemfile(char*);
    IGN(p);
    return (ldmemfile(filnam));
}

#ifdef __cplusplus
};
#endif

