                                                        /*    DL_OPCODES.C  */
#include "cs.h"

#if defined(WIN32) || defined(__unix) || defined(LINUX)
#define HAVE_DIRENT_H   1
#endif

#ifdef HAVE_DIRENT_H
#include <sys/types.h>
#include <dirent.h>
#endif

int csoundLoadExternal(void *csound, const char *libraryPath)
{
    void    *handle;
    void    *(*init)(GLOBALS*);
    long    (*size)(void);
    int     nr_opcodes = 0, length, olength;
    void    *opcodlst_n;

    handle = (void*) csoundOpenLibrary(libraryPath);
    if (handle == NULL) {
      /* could not open library */
      return (int) CSOUND_ERROR;
    }
    /* check the number of opcodes, */
    size = (long(*)(void)) csoundGetLibrarySymbol(handle, "opcode_count");
    if ((void*) size == NULL) {
      /* or the total size of OENTRY structures (deprecated) */
      size = (long(*)(void)) csoundGetLibrarySymbol(handle, "opcode_size");
    }
    else
      nr_opcodes = 1;         /* using newer interface */
    if ((void*) size == NULL) {
      /* still not found */
      err_printf(Str("Failed to initialise %s: no opcode count was found\n"),
                 libraryPath);
      return (int) CSOUND_ERROR;
    }
    length = (int) (*size)();
    if (!nr_opcodes) {
      if ((int) ((length / sizeof(OENTRY)) * sizeof(OENTRY)) != length) {
        err_printf(Str("%s is not compatible with this version of Csound\n"),
                   libraryPath);
        return (int) CSOUND_ERROR;
      }
    }
    if (length < 0) {
      err_printf(Str("%s: invalid opcode count: %d\n"), libraryPath, length);
      return (int) CSOUND_ERROR;
    }
    /* first check for an init function with the appropriate float type */
#ifdef MYFLT_IS_FLOAT
    init = (void *(*)(GLOBALS*))
      csoundGetLibrarySymbol(handle, "opcode_init_32");
#else
    init = (void *(*)(GLOBALS*))
      csoundGetLibrarySymbol(handle, "opcode_init_64");
#endif
    /* if that failed, fall back to the generic initialisation function */
    if ((void*) init == NULL) {
      init = (void *(*)(GLOBALS*))
        csoundGetLibrarySymbol(handle, "opcode_init");
    }
    /* still not found */
    if ((void*) init == NULL)  {
      err_printf(Str("Failed to initialise %s: no init function found\n"),
                 libraryPath);
      return (int) CSOUND_ERROR;
    }
    /* call init function */
    opcodlst_n = (*init)(&cglob);
    /* if no opcodes are to be allocated, done now */
    if (!length)
      return (int) CSOUND_SUCCESS;

    /* original number of opcodes */
    olength = oplstend - opcodlst;

    /* newer interface returns OENTRY** */
    if (nr_opcodes) {
      int i = 0;
      nr_opcodes = length;
      opcodlst = (OENTRY*) mrealloc(opcodlst,
                                    (olength + nr_opcodes) * sizeof(OENTRY));
      oplstend = opcodlst + olength;
      do {
        if (((OENTRY**) opcodlst_n)[i] != NULL) {
          /* clear entry */
          memset(oplstend, 0, sizeof(OENTRY));
          /* copy all the standard fields */
          oplstend->opname    = ((OENTRY**) opcodlst_n)[i]->opname;
          oplstend->dsblksiz  = ((OENTRY**) opcodlst_n)[i]->dsblksiz;
          oplstend->thread    = ((OENTRY**) opcodlst_n)[i]->thread;
          oplstend->outypes   = ((OENTRY**) opcodlst_n)[i]->outypes;
          oplstend->intypes   = ((OENTRY**) opcodlst_n)[i]->intypes;
          oplstend->iopadr    = ((OENTRY**) opcodlst_n)[i]->iopadr;
          oplstend->kopadr    = ((OENTRY**) opcodlst_n)[i]->kopadr;
          oplstend->aopadr    = ((OENTRY**) opcodlst_n)[i]->aopadr;
          oplstend->dopadr    = ((OENTRY**) opcodlst_n)[i]->dopadr;
          oplstend++;
        }
      } while (++i < nr_opcodes);
    }
    else {
      /* old interface: returns OENTRY* */
      opcodlst = (OENTRY*) mrealloc(opcodlst,
                                    olength * sizeof(OENTRY) + length);
      oplstend = opcodlst + olength;
      memcpy(oplstend, (OENTRY*) opcodlst_n, length);
      oplstend += (int) (length / sizeof(OENTRY));
    }
    return (int) CSOUND_SUCCESS;
}

int csoundLoadExternals(void *csound)
{
    char buffer[4096], dir_name[256], *plugins[256], *s, *s0;
    int  cnt = 0;

    if (((GLOBALS*) csound)->oplibs == NULL)
      buffer[0] = '\0';             /* no plugins specified on command line */
    else
      strcpy(buffer, ((GLOBALS*) csound)->oplibs);
    /* parse command line */
    s0 = &(buffer[0]);
    s = s0 - 1;
    while (*++s != '\0') {
      if (*s == ',') {
        *s = '\0';      /* terminate string */
        if ((int) strlen(s0) > 0)
          plugins[cnt++] = s0;
        s0 = s + 1;
      }
    }
    /* check for last option */
    if ((int) ((char*) s - (char*) s0) > 0)
      plugins[cnt++] = s0;
#ifdef HAVE_DIRENT_H
    /* on systems where it is supported, check the contents of 'OPCDIR' */
    /* directory, and load all libraries there */
    {
      DIR   *dirptr;
      char  *dirnam;
      struct dirent *ent;
      int   i;

      dirnam = ((CSOUNDAPI*) csound)->GetEnv("OPCDIR");
      if (dirnam == NULL) goto nolibdir;
      dirptr = opendir(dirnam);
      if (dirptr == NULL) goto nolibdir;
      /* directory name */
      strcpy((char*) dir_name, dirnam);
      i = strlen(dir_name);
      if (i > 0) {
        /* strip any trailing pathname delimiter */
        if (dir_name[i - 1] == DIRSEP || dir_name[i - 1] == '/') i--;
        /* and now add one to make sure there is exactly one */
        dir_name[i++] = DIRSEP;
        dir_name[i] = '\0';
      }
      while ((ent = readdir(dirptr)) != NULL) {
        /* if found file, */
        s0 = (char*) ent->d_name;
        i = strlen(s0);
#ifdef WIN32
        /* with the right extension, */
        if (i > 4 &&
            (s0[i - 1] == 'L' || s0[i - 1] == 'l') &&
            (s0[i - 2] == 'L' || s0[i - 2] == 'l') &&
            (s0[i - 3] == 'D' || s0[i - 3] == 'd') &&
            s0[i - 4] == '.') {
#else
        if (i > 3 &&
            (s0[i - 1] == 'O' || s0[i - 1] == 'o') &&
            (s0[i - 2] == 'S' || s0[i - 2] == 's') &&
            s0[i - 3] == '.') {
#endif
          /* add to list of libraries to load */
          s++;
          strcpy(s, dir_name);  /* with full path */
          strcat(s, s0);
          plugins[cnt++] = s;
          i = strlen(s);
          s += i;
        }
      }
      closedir(dirptr);
    }
 nolibdir:
#endif          /* HAVE_DIRENT_H */
    plugins[cnt] = NULL;

    if (cnt == 0)                       /* nothing to load */
      return CSOUND_SUCCESS;

    err_printf(Str("Loading libraries ...\n"));
    for (cnt = 0; plugins[cnt] != NULL; cnt++) {
      err_printf("  %s\n", plugins[cnt]);
      csoundLoadExternal(csound, plugins[cnt]);
    }

    return CSOUND_SUCCESS;
}

