#ifndef CSOUND_H
#define CSOUND_H
/**
* S I L E N C E
*
* An auto-extensible system for making music on computers by means of
* software alone.
* Copyright (c) 2001 by Michael Gogins. All rights reserved.
*
* L I C E N S E
*
* This software is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* P U R P O S E
*
* Declares the public C application programming interface to Csound.
*/
#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

/* This header file can be used in three different ways controlled by the */
/* macros CSOUND_PLUGIN_LIBRARY and CSOUND_API_HEADERS.                   */
/*   1. For use from main Csound sources (by including "cs.h"), none of   */
/*      the above macros are defined (this allows access to all internal  */
/*      structures and functions that may change between releases).       */
/*   2. For CsoundAPI applications (executables), define                  */
/*      CSOUND_API_HEADERS only. This limits access to only the standard  */
/*      structures and library functions. Using this interface requires   */
/*      including "Csound/csound.h" (this file), and defining             */
/*      CSOUND_API_HEADERS. Additionally, many header files from the      */
/*      Csound sources, and the Csound library are needed for compiling   */
/*      and linking the application.                                      */
/*   3. Plugin libraries (.so or DLL files) should be compiled with       */
/*      CSOUND_PLUGIN_LIBRARY (which implies CSOUND_API_HEADERS).         */
/*      The difference compared to CsoundAPI applications is that no      */
/*      functions are declared (the only way to access the API is through */
/*      the function pointers in the CSOUNDAPI structure passed to the    */
/*      plugin as 'void *csound'). Compiling plugins requires including   */
/*      "Csound/csdl.h"; the only files needed for compilation are        */
/*      "Csound/csdl.h", "Csound/csound.h", "Csound/csstruct.h", and      */
/*      "Csound/sysdep.h", and no libraries.                              */

#ifdef CSOUND_PLUGIN_LIBRARY
#define CSOUND_API_HEADERS  1
#endif

#include "Csound/sysdep.h"
#include <stdarg.h>
#include "Csound/csstruct.h"

#ifndef CSOUND_API_HEADERS
#include "cwindow.h"
#include "opcode.h"
#else
typedef void    WINDAT;     /* stub */
#endif

        /**
        * ERROR DEFINITIONS
        */

        typedef enum
        {
                CSOUND_SUCCESS        =  0,
                CSOUND_ERROR          = -1,
                CSOUND_INITIALIZATION = -2,
                CSOUND_PERFORMANCE    = -3,
                CSOUND_MEMORY         = -4,
                CSOUND_AUXALLOC_ERROR = -5
        }
        CSOUND_STATUS;

/* -- standard Csound globals that can be accessed from user applications -- */

/* WARNING: do not change this structure ! */

typedef struct CSGLOBALS_S {
    int         ksmps;              /* ksmps of current instrument instance  */
    int         global_ksmps;       /* orchestra level ksmps                 */
    MYFLT       ensmps;             /* ksmps as float/double                 */
    MYFLT       global_ensmps;      /* global_ksmps as float/double          */
    MYFLT       ekr;                /* control rate of curr. instr. instance */
    MYFLT       global_ekr;         /* orchestra level control rate          */
    MYFLT       onedkr;             /* FL(1.0) / ekr                         */
    MYFLT       global_onedkr;      /* FL(1.0) / global_ekr                  */
    MYFLT       hfkprd;             /* FL(0.5) / ekr                         */
    MYFLT       global_hfkprd;      /* FL(0.5) / global_ekr                  */
    MYFLT       kicvt;              /* FMAXLEN / ekr                         */
    MYFLT       global_kicvt;       /* FMAXLEN / global_ekr                  */
    long        kcounter;           /* number of k-cycles in onedkr units    */
    long        global_kcounter;    /* nr of k-cycles in global_onedkr units */
    MYFLT       esr;                /* sample rate                           */
    MYFLT       onedsr;             /* FL(1.0) / esr                         */
    MYFLT       sicvt;              /* FMAXLEN / esr                         */
    MYFLT       tpidsr;             /* 2PI / esr                             */
    MYFLT       mpidsr;             /* -PI / esr                             */
    MYFLT       mtpdsr;             /* -2PI / esr                            */
    MYFLT       e0dbfs;             /* 0dBFS level (32767 by default)        */
    MYFLT       __dummy_01;         /* unused                                */
    int         nchnls;             /* number of audio channels              */
    int         nspin;              /* global_ksmps * nchnls                 */
    int         nspout;             /* global_ksmps * nchnls                 */
    int         spoutactive;        /* non-zero if already wrote to spout    */
    MYFLT       * spin;             /* float audio in buffer (nspin samples) */
    MYFLT       * spout;            /* float audio out buffer (nspout smpls) */
    long        zklast;             /* highest k-rate ZAK location           */
    long        zalast;             /* highest a-rate ZAK location           */
    MYFLT       * zkstart;          /* k-rate ZAK space (NULL if none)       */
    MYFLT       * zastart;          /* a-rate ZAK space (NULL if none)       */
    INSDS       * curip;            /* curr. instr. instance (i-time only !) */
    OPDS        * ids;              /* current init opcode (i-time only)     */
    OPDS        * pds;              /* current perf opcode (perf time only)  */
    int         * msglevel;         /* pointer to message level (-m) value   */
    int         reinitflag;         /* non-zero in reinit mode               */
    int         tieflag;            /* non-zero for tied notes               */
    struct CSOUNDAPI_S  * csound;   /* pointer to API functions              */
    char        ** strsets;         /* pointer to array of strset strings    */
    int         strsmax;            /* maximum strset array index            */
    int         __dummy_02;         /* unused                                */
} CSGLOBALS;

/* --------------------- Csound API functions ---------------------- */
/* can be accessed by casting the 'void *csound' instance pointer to */
/* CSOUNDAPI* (e.g. ((CSOUNDAPI*) csound)->GetVersion())             */

typedef void (*dummyFunc_t)(void);

typedef struct CSOUNDAPI_S {
    /* ---- General functions ---- */
    int     (*GetVersion)(void);
    int     (*GetAPIVersion)(void);
    void    *(*GetHostData)(void *csound);
    void    (*SetHostData)(void *csound, void *hostData);
    int     (*Perform)(void *csound, int argc, char **argv);
    int     (*Compile)(void *csound, int argc, char **argv);
    int     (*PerformKsmps)(void *csound);
    int     (*PerformBuffer)(void *csound);
    void    (*Cleanup)(void *csound);
    void    (*Reset)(void *csound);
    /* ---- Attributes ---- */
    CSGLOBALS   *(*GetGlobals)(void *csound);
    MYFLT   (*GetSr)(void *csound);
    MYFLT   (*GetKr)(void *csound);
    MYFLT   (*GetGlobalKr)(void *csound);
    int     (*GetKsmps)(void *csound);
    int     (*GetGlobalKsmps)(void *csound);
    int     (*GetNchnls)(void *csound);
    MYFLT   (*Get0dBFS)(void *csound);
    int     (*GetSampleFormat)(void *csound);
    int     (*GetSampleSize)(void *csound);
    long    (*GetInputBufferSize)(void *csound);
    long    (*GetOutputBufferSize)(void *csound);
    void    *(*GetInputBuffer)(void *csound);
    void    *(*GetOutputBuffer)(void *csound);
    MYFLT   *(*GetSpin)(void *csound);
    MYFLT   *(*GetSpout)(void *csound);
    MYFLT   (*GetScoreTime)(void *csound);
    MYFLT   (*GetProgress)(void *csound);
    MYFLT   (*GetProfile)(void *csound);
    MYFLT   (*GetCpuUsage)(void *csound);
    /* ---- Score ---- */
    int     (*IsScorePending)(void *csound);
    void    (*SetScorePending)(void *csound, int pending);
    MYFLT   (*GetScoreOffsetSeconds)(void *csound);
    void    (*SetScoreOffsetSeconds)(void *csound, MYFLT offset);
    void    (*RewindScore)(void *csound);
    /* ---- Messages ---- */
    void    (*Message)(void *csound, const char *format, ...);
    void    (*MessageV)(void *csound, const char *format, va_list args);
    void    (*MessageS)(void *csound, const char *format, va_list args);
    void    (*ThrowMessage)(void *csound, const char *format, ...);
    void    (*ThrowMessageV)(void *csound, const char *format, va_list args);
    void    (*SetMessageCallback)(void *csound,
                                  void (*csoundMessageCallback)
                                       (void *hostData,
                                        const char *format, va_list valist));
    void    (*SetMessageVCallback)(void *csound,
                                   void (*csoundMessageCallback)
                                        (void *hostData,
                                         const char *format, va_list valist));
    void    (*SetMessageSCallback)(void *csound,
                                   void (*csoundMessageCallback)
                                        (void *hostData,
                                         const char *format, va_list valist));
    void    (*SetThrowMessageCallback)(void *csound,
                                       void (*throwMessageCallback)
                                            (void *hostData,
                                             const char *format,
                                             va_list valist));
    int     (*GetMessageLevel)(void *csound);
    void    (*SetMessageLevel)(void *csound, int messageLevel);
    void    (*InputMessage)(void *csound, const char *message);
    /* ---- Sensing and control ---- */
    void    (*KeyPress)(void *csound, char c);
    char    (*GetChar)(void);
    void    (*SetInputValueCallback)(void *csound,
                                     void (*inputValueCalback)
                                          (void *hostData, char *channelName,
                                           MYFLT *value));
    void    (*InputValue)(char *channelName, MYFLT *value);
    void    (*SetOutputValueCallback)(void *csound,
                                      void (*outputValueCalback)
                                           (void *hostData, char *channelName,
                                            MYFLT value));
    void    (*OutputValue)(char *channelName, MYFLT value);
    void    (*ScoreEvent)(void *csound, char type, MYFLT *pFields,
                          long numFields);
    /* ---- MIDI ---- */
    void    (*SetExternalMidiOpenCallback)(void *csound,
                                           void (*midiOpenCallback)
                                                (void *hostData));
    void    (*SetExternalMidiReadCallback)(void *csound,
                                           int (*readMidiCallback)
                                               (void *hostData,
                                                unsigned char *midiData,
                                                int size));
    void    (*SetExternalMidiWriteCallback)(void *csound,
                                            int (*writeMidiCallback)
                                                (void *hostData,
                                                 unsigned char *midiData));
    void    (*SetExternalMidiCloseCallback)(void *csound,
                                            void (*closeMidiCallback)
                                                 (void *hostData));
    int     (*IsExternalMidiEnabled)(void *csound);
    void    (*SetExternalMidiEnabled)(void *csound, int enabled);
    /* ---- display/graphics ---- */
    void    (*SetIsGraphable)(void *csound, int isGraphable);
    void    (*SetMakeGraphCallback)(void *csound,
                                    void (*makeGraphCallback)(void *hostData,
                                                              WINDAT *p,
                                                              char *name));
    void    (*SetDrawGraphCallback)(void *csound,
                                    void (*drawGraphCallback)
                                         (void *hostData, WINDAT *p));
    void    (*SetKillGraphCallback)(void *csound,
                                    void (*killGraphCallback)(void *hostData,
                                                              WINDAT *p));
    void    (*SetExitGraphCallback)(void *csound, int (*exitGraphCallback)
                                                      (void *hostData));
    /* ---- Opcodes ---- */
    opcodelist  *(*NewOpcodeList)(void);
    void    (*DisposeOpcodeList)(opcodelist *opcodelist_);
    int     (*AppendOpcode)(void *csound,
                            char *opname, int dsblksiz, int thread,
                            char *outypes, char *intypes, SUBR iopadr,
                            SUBR kopadr, SUBR aopadr, SUBR dopadr);
    /* ---- Libraries ---- */
    int     (*LoadExternal)(void *csound, const char *libraryPath);
    int     (*LoadExternals)(void *csound);
    void    *(*OpenLibrary)(const char *libraryPath);
    void    *(*CloseLibrary)(void *library);
    void    *(*GetLibrarySymbol)(void *library, const char *procedureName);
    /* ---- Misc. functions ---- */
    void    (*SetYieldCallback)(void *csound, int (*yieldCallback)
                                                  (void *hostData));
    void    (*SetEnv)(void *csound, const char *environmentVariableName,
                      const char *path);
    char    *(*GetEnv)(const char *name);
    /* ---- Real-time audio ---- */
    void    (*SetPlayopenCallback)(void *csound,
                                   void *(*playopen__)(void *csound,
                                                       AUDIOPARMS *devParms));
    void    (*SetRtplayCallback)(void *csound,
                                 void (*rtplay__)(void *audioOutData,
                                                  char *outBuf, int nbytes));
    void    (*SetRecopenCallback)(void *csound,
                                  void *(*recopen__)(void *csound,
                                                     AUDIOPARMS *devParms));
    void    (*SetRtrecordCallback)(void *csound,
                                   int (*rtrecord__)(void *audioInData,
                                                     char *inBuf, int nbytes));
    void    (*SetRtcloseCallback)(void *csound, void (*rtclose__)
                                                     (void *audioInData,
                                                      void *audioOutData));
    /* ---- Error handling ---- */
    void    (*SetInitErrorCallback)(void *csound,
                                    void (*initErrorCallback)
                                         (void*, char*));
    void    (*InitError)(void *csound, char *s);
    void    (*SetPerfErrorCallback)(void *csound,
                                    void (*perfErrorCallback)
                                         (void*, char*));
    void    (*PerfError)(void *csound, char *s);
    void    (*SetDieCallback)(void *csound,
                              void (*dieCallback)(void*, char*));
    void    (*Die)(void *csound, char *s);
    void    (*SetErrorHandlerCallback)(void *csound,
                                       void (*errorHandlerCallback)
                                            (void*, int, char*, void*));
    void    (*Error)(void *csound, int errCode, char *errMsg, void *userData);
    /* ---- Opcode helper functions ---- */
    void    *(*GetEngineForOpcode)(void *p);
    void    *(*GetGlobalsForOpcode)(void *p);
    void    (*GetOpcodeParams)(void *p, int *outocount, int *xoutcode,
                                        int *inocount,  int *xincode,
                                        char ***strargs);
    MCHNBLK *(*GetMIDIParams)(void *p);
    FUNC    *(*QueryFTableInit)(void *p, MYFLT *argp);
    FUNC    *(*QueryFTablePerf)(void *p, MYFLT *argp);
    FUNC    *(*QueryDeferredFTableInit)(void *p, MYFLT *argp);
    MYFLT   (*GetReleaseTime)(void *p);
    void    (*SetReleaseTime)(void *p, MYFLT xtim);
    int     (*GetReleaseFlag)(void *p);
    int     (*GetBeatModeFlag)(void *p);
    MYFLT   (*GetOffTime)(void *p);
    void    (*SetOffTime)(void *p, MYFLT offtim);
    int     (*GetPFields)(void *p, MYFLT **pflds);
    void    (*SkipInitOpcodes)(void *p);
    void    (*SkipPerfOpcodes)(void *p);
    void    (*TurnOffCurrentNote)(void *p);
    void    (*AuxAlloc)(void *p, AUXCH *auxchp, size_t nbytes);
    int     (*OpenFile)(void *p, FDCH *fdchp, const char *name, int mode);
    void    (*CloseFile)(void *p, FDCH *fdchp);
    MEMFIL  *(*LoadFile)(void *p, char *filnam);
    MYFLT   (*RandomValueU)(void *p);
    MYFLT   (*RandomValueB)(void *p);
    void    (*SetRandomSeed)(void *p, unsigned int seedValue);
    /* ---- Time functions ---- */
    void    (*timers_struct_init)(RTCLOCK*);
    double  (*timers_get_real_time)(RTCLOCK*);
    double  (*timers_get_CPU_time)(RTCLOCK*);
    unsigned long   (*timers_random_seed)(void);
    /* ---- Memory management ---- */
    void    *(*HeapCreate)(void *csound, int flags);
    void    (*HeapSetErrorCode)(void *handle, int errCode);
    void    *(*Malloc)(void *handle, size_t nbytes);
    void    *(*Calloc)(void *handle, size_t nmemb, size_t size);
    void    (*Free)(void *handle, void *ptr);
    void    *(*Realloc)(void *handle, void *ptr, size_t nbytes);
    void    (*HeapDestroy)(void **handle);
    /* ---- Native language support ---- */
    char    *(*LocalizeString)(const char *s);
} CSOUNDAPI;

/* -------------------- API functions and documentation -------------------- */

#define MEMALLOC_ERROR_RETURN   0x00000001
#define MEMALLOC_DISPOSABLE     0x00000002

/* plugin DLLs should use the function pointers in the CSOUNDAPI structure */

#ifndef CSOUND_PLUGIN_LIBRARY

        /*
        * INSTANTIATION
        */

        /**
        * Creates an instance of Csound.
        * Returns an opaque pointer that must be passed to most Csound API
        * functions.
        * The hostData parameter can be null, or it can be a pointer to any
        * sort of data; this pointer can be accessed from the Csound instance
        * that is passed to callback routines.
        */
        void *csoundCreate(void *hostData);

        /**
        * Returns a pointer to the requested interface, if available, in the
        * interface argument, and its version number, in the version argument.
        * Returns 0 for success and 1 for failure.
        */
        int csoundQueryInterface(const char *name, void **iface, int *version);

        /**
        * Destroys an instance of Csound.
        */
        void csoundDestroy(void *csound);

        /**
        * Returns the version number times 100 (4.20 = 420).
        */
        int csoundGetVersion(void);

        /*
        * Returns the API version number times 100 (1.10 = 110).
        */
        int csoundGetAPIVersion(void);

        /**
        * Returns host data.
        */
        void *csoundGetHostData(void *csound);

        /**
        * Sets host data.
        */
        void csoundSetHostData(void *csound, void *hostData);

        /*
        * PERFORMANCE
        */

        /**
        * Compiles and renders a Csound performance,
        * as directed by the supplied command-line arguments,
        * in one pass. Returns 1 for success, 0 for failure.
        */
        int csoundPerform(void *csound, int argc, char **argv);

        /**
        * Compiles Csound input files (such as an orchestra and score)
        * as directed by the supplied command-line arguments,
        * but does not perform them. Returns a non-zero error code on failure.
        * In this (host-driven) mode, the sequence of calls should be as
        * follows:
        *
        *       csoundCompile(argc, argv, thisObj);
        *       while(!csoundPerformBuffer());
        *       csoundCleanup();
        *       csoundReset();
        */
        int csoundCompile(void *csound, int argc, char **argv);

        /**
        * Senses input events, and performs one control sample worth
        * (ksmps) of audio output.
        * Note that csoundCompile must be called first.
        * Returns false during performance, and true when performance is
        * finished.
        * If called until it returns true, will perform an entire score.
        * Enables external software to control the execution of Csound,
        * and to synchronize performance with audio input and output.
        */
        int csoundPerformKsmps(void *csound);

        /**
        * Performs Csound, sensing real-time and score events
        * and processing one buffer's worth (-b frames) of interleaved audio.
        * Returns a pointer to the new output audio in 'outputAudio'
        * Note that csoundCompile must be called first, then call
        * csoundGetOutputBuffer()
        * and csoundGetInputBuffer() to get the pointer to csound's i/o buffers.
        * Returns false during performance, and true when performance is
        * finished.
        */
        int csoundPerformBuffer(void *csound);

        /**
        * Prints information about the end of a performance.
        * Must be called after the final call to csoundPerformKsmps.
        */
        void csoundCleanup(void *csound);

        /**
        * Resets all internal memory and state in preparation for a new
        * performance.
        * Enables external software to run successive Csound performances
        * without reloading Csound.
        */
        void csoundReset(void *csound);

        /*
        * ATTRIBUTES
        */

        /**
        * Returns pointer to a structure with commonly used globals
        * (see above for structure definition).
        */
        CSGLOBALS *csoundGetGlobals(void *csound);

        /**
        * Returns the number of audio sample frames per second.
        */
        MYFLT csoundGetSr(void *csound);

        /**
        * Returns the number of control samples per second in
        * the currently performing instrument instance.
        */
        MYFLT csoundGetKr(void *csound);

        /**
        * Returns the number of control samples per second as
        * defined in the orchestra header.
        */
        MYFLT csoundGetGlobalKr(void *csound);

        /**
        * Returns the number of audio sample frames per control sample
        * in the currently performing instrument instance.
        */
        int csoundGetKsmps(void *csound);

        /**
        * Returns the number of audio sample frames per control sample
        * as defined in the orchestra header.
        */
        int csoundGetGlobalKsmps(void *csound);

        /**
        * Returns the number of audio output channels.
        */
        int csoundGetNchnls(void *csound);

        /**
        * Returns the full scale 0dB level (usually 32767.0)
        */
        MYFLT csoundGet0dBFS(void *csound);

        /**
        * Returns the sample format.
        */
        int csoundGetSampleFormat(void *csound);

        /**
        * Returns the size in bytes of a single sample.
        */
        int csoundGetSampleSize(void *csound);

        /**
        * Returns the number of samples in Csound's input buffer.
        */
        long csoundGetInputBufferSize(void *csound);

        /**
        * Returns the number of samples in Csound's output buffer.
        */
        long csoundGetOutputBufferSize(void *csound);

        /**
        * Returns the address of the Csound audio input buffer.
        * Enables external software to write audio into Csound before calling
        * csoundPerformBuffer
        */
        void *csoundGetInputBuffer(void *csound);

        /**
        * Returns the address of the Csound audio output buffer.
        * Enables external software to read audio from Csound after calling
        * csoundPerformBuffer.
        */
        void *csoundGetOutputBuffer(void *csound);

        /**
        * Returns the address of the Csound audio input working buffer (spin).
        * Enables external software to write audio into Csound before calling
        * csoundPerformKsmps.
        */
        MYFLT *csoundGetSpin(void *csound);

        /**
        * Returns the address of the Csound audio output working buffer (spout).
        * Enables external software to read audio from Csound after calling
        * csoundPerformKsmps.
        */
        MYFLT *csoundGetSpout(void *csound);

        /**
        * Returns the current score time.
        */
        MYFLT csoundGetScoreTime(void *csound);

        /**
        * Returns the % of score completed.
        */
        MYFLT csoundGetProgress(void *csound);

        /**
        * Returns the scoreTime vs. calculatedTime ratio.  For real-time
        * performance this value should be always == 1.
        */
        MYFLT csoundGetProfile(void *csound);

        /**
        * Returns the sampsTime vs. calculatedTime ratio.
        */
        MYFLT csoundGetCpuUsage(void *csound);

        /*
        * SCORE HANDLING
        */

        /**
        * Returns whether Csound's score is synchronized with external software.
        */
        int csoundIsScorePending(void *csound);

        /**
        * Sets whether Csound's score is synchronized with external software.
        */
        void csoundSetScorePending(void *csound, int pending);

        /**
        * Csound events prior to the offset are consumed and discarded prior
        * to beginning performance.
        * Can be used by external software to begin performance midway
        * through a Csound score.
        */
        MYFLT csoundGetScoreOffsetSeconds(void *csound);

        /**
        * Csound events prior to the offset are consumed and discarded prior
        * to beginning performance.
        * Can be used by external software to begin performance midway
        * through a Csound score.
        */
        void csoundSetScoreOffsetSeconds(void *csound, MYFLT offset);

        /**
        * Rewinds a compiled Csound score to its beginning.
        */
        void csoundRewindScore(void *csound);

        /*
        * MESSAGES & TEXT
        */

        /**
        * Displays an informational message.
        */
        void csoundMessage(void *csound, const char *format, ...);
        void csoundMessageV(void *csound, const char *format, va_list args);
        void csoundMessageS(void *csound, const char *format, va_list args);

        /**
        * Throws an informational message as a C++ exception.
        */
        void csoundThrowMessage(void *csound, const char *format, ...);
        void csoundThrowMessageV(void *csound, const char *format,
                                 va_list args);

        /**
        * Sets a function to be called by Csound to print an informational
        * message.
        */
        void csoundSetMessageCallback(void *csound,
                                      void (*csoundMessageCallback)
                                           (void *csound, const char *format,
                                            va_list valist));
        void csoundSetMessageVCallback(void *csound,
                                       void (*csoundMessageCallback)
                                            (void *csound, const char *format,
                                             va_list valist));
        void csoundSetMessageSCallback(void *csound,
                                       void (*csoundMessageCallback)
                                            (void *csound, const char *format,
                                             va_list valist));

        /**
        * Sets a funtion for Csound to stop execution with an error message
        * or exception.
        */
        void csoundSetThrowMessageCallback(void *csound,
                                           void (*throwMessageCallback)
                                                (void *csound,
                                                 const char *format,
                                                 va_list valist));

        /**
        * Returns the Csound message level (from 0 to 7).
        */
        int csoundGetMessageLevel(void *csound);

        /**
        * Sets the Csound message level (from 0 to 7).
        */
        void csoundSetMessageLevel(void *csound, int messageLevel);

        /**
        * Input a NULL-terminated string (as if from a console) usually
        * used for lineevents
        */
        void csoundInputMessage(void *csound, const char *message);

        /**
        * Set the ascii code of most recent key pressed.
        * This value is used by the 'keypress' opcode.
        */
        void csoundKeyPress(void *csound, char c);

        /**
        * Returns the ASCII code of the most recently pressed key
        * (or the most recent character set by csoundKeyPress()).
        */
        char csoundGetChar(void);

        /*
        * CONTROL AND EVENTS
        */

        /**
        * Control values are specified by a 'channelName' string.
        * Note that the 'invalue' & 'outvalue' channels can be specified by
        * either a string or a number.  If a number is specified, it will be
        * converted to a string before making the callbacks to the external
        * software.
        * numerical channel names into text, so if the orchestra contains
        */

        /**
        * Called by external software to set a function for Csound to fetch
        * input control values.
        * The 'invalue' opcodes will directly call this function.
        */
        void csoundSetInputValueCallback(void *csound,
                                         void (*inputValueCalback)
                                              (void *csound, char *channelName,
                                               MYFLT *value));

        /**
        * Read an input control value using the callback function
        * set by csoundSetInputValueCallback().
        */
        void csoundInputValue(char *channelName, MYFLT *value);

        /**
        * Called by external software to set a function for Csound to send
        * output control values.
        * The 'outvalue' opcodes will directly call this function.
        */
        void csoundSetOutputValueCallback(void *csound,
                                          void (*outputValueCalback)
                                               (void *csound, char *channelName,
                                                MYFLT value));

        /**
        * Send an output control value using the callback function
        * set by csoundSetOutputValueCallback().
        */
        void csoundOutputValue(char *channelName, MYFLT value);

        /**
        * Send a new score event. 'type' is the score event type
        * ('i', 'f', or 'e')
        * 'numFields' is the size of the pFields array.  'pFields' is an array
        * of floats with all the pfields for this event, starting with the p1
        * value specified in pFields[0].
        */
        void csoundScoreEvent(void *csound, char type, MYFLT *pFields,
                              long numFields);

        /*
        * MIDI
        */

        /**
        * Called by external software to set a function for Csound to call
        * to open MIDI input.
        */
        void csoundSetExternalMidiOpenCallback(void *csound,
                                               void (*midiOpenCallback)
                                                    (void *csound));

        /**
        * Called by external software to set a function for Csound to call
        * to read MIDI messages.
        */
        void csoundSetExternalMidiReadCallback(void *csound,
                                               int (*readMidiCallback)
                                                   (void *csound,
                                                    unsigned char *midiData,
                                                    int size));

        /**
        * Called by external software to set a function for Csound to call
        * to write a 4-byte MIDI message.
        */
        void csoundSetExternalMidiWriteCallback(void *csound,
                                                int (*writeMidiCallback)
                                                    (void *csound,
                                                     unsigned char *midiData));

        /**
        * Called by external software to set a function for Csound to call
        * to close MIDI input.
        */
        void csoundSetExternalMidiCloseCallback(void *csound,
                                                void (*closeMidiCallback)
                                                     (void *csound));

        /**
        * Returns true if external MIDI is enabled, and false otherwise.
        */
        int csoundIsExternalMidiEnabled(void *csound);

        /**
        * Sets whether external MIDI is enabled.
        */
        void csoundSetExternalMidiEnabled(void *csound, int enabled);

        /*
        * FUNCTION TABLE DISPLAY
        */

        /**
        * Tells Csound supports external graphic table display.
        */
        void csoundSetIsGraphable(void *csound, int isGraphable);

        /**
        * Called by external software to set Csound's MakeGraph function.
        */
        void csoundSetMakeGraphCallback(void *csound,
                                        void (*makeGraphCallback)
                                             (void *csound, WINDAT *windat,
                                              char *name));

        /**
        * Called by external software to set Csound's DrawGraph function.
        */
        void csoundSetDrawGraphCallback(void *csound,
                                        void (*drawGraphCallback)
                                             (void *csound, WINDAT *windat));

        /**
        * Called by external software to set Csound's KillGraph function.
        */
        void csoundSetKillGraphCallback(void *csound,
                                        void (*killGraphCallback)
                                             (void *csound, WINDAT *windat));

        /**
        * Called by external software to set Csound's ExitGraph function.
        */
        void csoundSetExitGraphCallback(void *csound,
                                        int (*exitGraphCallback)(void *csound));

        /*
        * OPCODES
        */

        /*
        * Gets a list of all opcodes.
        * Make sure to call csoundDisposeOpcodeList() when done with the list.
        */
        opcodelist *csoundNewOpcodeList(void);

        /*
        * Releases an opcode list
        */
        void csoundDisposeOpcodeList(opcodelist *opcodelist_);

        /**
        * Appends an opcode implemented by external software
        * to Csound's internal opcode list.
        * The opcode list is extended by one slot,
        * and the parameters are copied into the new slot.
        */
        int csoundAppendOpcode(void *csound,
                               char *opname, int dsblksiz, int thread,
                               char *outypes, char *intypes,
                               void (*iopadr)(void *),
                               void (*kopadr)(void *),
                               void (*aopadr)(void *),
                               void (*dopadr)(void *));

        /*
        * Signature for external registration function,
        * such as for plugin opcodes or scripting languages.
        * The external has complete access to the Csound API,
        * so a plugin opcode can just call csound->AppendOpcode()
        * for each of its opcodes.
        */
        typedef PUBLIC int (*CsoundRegisterExternalType)(void *csound);

        /*
        * Registers all opcodes in the library.
        */
        int csoundLoadExternal(void *csound, const char *libraryPath);

        /*
        * Registers all opcodes in all libraries in the opcodes directory.
        */
        int csoundLoadExternals(void *csound);

        /*
        * MISCELLANEOUS FUNCTIONS
        */

        /*
        * Platform-independent function
        * to load a shared library.
        */
        void *csoundOpenLibrary(const char *libraryPath);

        /*
        * Platform-independent function
        * to unload a shared library.
        */
        void *csoundCloseLibrary(void *library);

        /*
        * Platform-independent function
        * to get a symbol address
        * in a shared library.
        */
        void *csoundGetLibrarySymbol(void *library, const char *symbolName);

        /**
        * Called by external software to set a function for
        * checking system events, yielding cpu time for
        * coopertative multitasking, etc..
        * This function is optional.  It is often used as a way
        * to 'turn off' Csound, allowing it to exit gracefully.
        * In addition, some operations like utility analysis
        * routines are not reentrant and you should use this
        * function to do any kind of updating during the operation.
        *
        * Returns an 'OK to continue' boolean
        */
        void csoundSetYieldCallback(void *csound,
                                    int (*yieldCallback)(void *csound));

        /**
        * Sets an evironment path for a getenv() call in Csound.
        * you can also use this method as a way to have different
        * csound instances have different default directories,
        * change the default dirs during performance, etc..
        *
        * Currently, Csound uses these 'envi' names only:
        *       "SSDIR", "SFDIR", "SADIR", "SFOUTYP", "INCDIR",
        *               "CSSTRNGS", "MIDIOUTDEV", and "HOME"
        */
        void csoundSetEnv(void *csound,
                          const char *environmentVariableName,
                          const char *path);

        /**
        * Get an environment variable set by csoundSetEnv() or
        * from the system.
        */
        char *csoundGetEnv(const char *name);

        /**
        *       REAL-TIME AUDIO PLAY AND RECORD
        */

        /**
        * Sets a function to be called by Csound for opening real-time audio
        * playback.
        */
        void csoundSetPlayopenCallback(void *csound,
                                       void *(*playopen__)
                                             (void *csound,
                                              AUDIOPARMS *devParms));

        /**
        * Sets a function to be called by Csound for performing real-time
        * audio playback.
        */
        void csoundSetRtplayCallback(void *csound,
                                     void (*rtplay__)
                                          (void *audioOutData,
                                           char *outBuf, int nbytes));

        /**
        * Sets a function to be called by Csound for opening real-time
        * audio recording.
        */
        void csoundSetRecopenCallback(void *csound,
                                      void *(*recopen__)
                                            (void *csound,
                                             AUDIOPARMS *devParms));

        /**
        * Sets a function to be called by Csound for performing real-time
        * audio recording.
        */
        void csoundSetRtrecordCallback(void *csound,
                                       int (*rtrecord__)
                                           (void *audioInData,
                                            char *inBuf, int nbytes));

        /**
        * Sets a function to be called by Csound for closing real-time
        * audio playback and recording.
        */
        void csoundSetRtcloseCallback(void *csound,
                                      void (*rtclose__)
                                           (void *audioInData,
                                            void *audioOutData));

        /**
        *
        *   ERROR HANDLING
        *
        */

        /**
        * Set init error callback function.
        */
        void csoundSetInitErrorCallback(void *csound,
                                        void (*initErrorCallback)
                                             (void*, char*));

        /**
        * Init error ('s' is the error message).
        */
        void csoundInitError(void *csound, char *s);

        /**
        * Set perf error callback function.
        */
        void csoundSetPerfErrorCallback(void *csound,
                                        void (*perfErrorCallback)
                                             (void*, char*));

        /**
        * Perf error ('s' is the error message).
        */
        void csoundPerfError(void *csound, char *s);

        /**
        * Set callback function for terminating a Csound instance
        * in case of a fatal error.
        */
        void csoundSetDieCallback(void *csound,
                                  void (*dieCallback)(void*, char*));

        /**
        * Terminate a Csound instance (called on fatal errors).
        * 's' is the error message.
        */
        void csoundDie(void *csound, char *s);

        /**
        * Set callback function for general error handler.
        */
        void csoundSetErrorHandlerCallback(void *csound,
                                           void (*errorHandlerCallback)
                                                (void*, int, char*, void*));

        /**
        * Call error handler.
        *   errCode     the error code (see also CSOUND_STATUS above)
        *   errMsg      the error message to be printed
        *   userData    pointer to data related to the error (this is
        *               unused in most cases)
        * csoundError(csound, CSOUND_INITIALIZATION, errMsg, NULL)
        *   is normally the same as csoundInitError(csound, errMsg), and
        * csoundError(csound, CSOUND_PERFORMANCE, errMsg, NULL)
        *   is equivalent to csoundPerfError(csound, errMsg)
        */
        void csoundError(void *csound, int errCode, char *errMsg,
                         void *userData);

        /**
        *
        * OPCODE HELPER FUNCTIONS
        *
        * (Note: in all these functions, 'p' is a pointer to the
        * data structure of the current opcode.)
        *
        */

        /**
        * Return pointer to GLOBALS structure of the Csound instance
        * owning the calling opcode (a simpler way to access this
        * structure is p->h.globals->csound).
        */
        void *csoundGetEngineForOpcode(void *p);

        /**
        * Return pointer to CSGLOBALS structure of the Csound instance
        * owning the calling opcode (a simpler way to access this
        * structure is p->h.globals), containing public data (such as
        * the sample rate, etc.) that can be used by the opcode.
        */
        void *csoundGetGlobalsForOpcode(void *p);

        /**
        * Get opcode parameters and store in variables pointed by
        * input args:
        *   outocount: number of output arguments
        *   xoutcode:  bitwise OR of
        *                1: first output argument is a-rate
        *                2: second output argument is a-rate
        *                4: third output argument is a-rate
        *                etc.
        *              xoutcode is set for the first 15 output arguments
        *   inocount:  number of input arguments
        *   xincode:   bitwise OR of
        *                1: first input argument is a-rate
        *                2: second input argument is a-rate
        *                4: third input argument is a-rate
        *                etc.
        *              xincode is set for the first 15 input arguments
        *   strargs:   pointer to an array of string arguments to the opcode
        *              Up to 4 string arguments are supported.
        *              An argument is a string if it has the value of the
        *              SSTRCOD macro.
        */
        void csoundGetOpcodeParams(void *p, int *outocount, int *xoutcode,
                                            int *inocount,  int *xincode,
                                            char ***strargs);

        /**
        * Return pointer to MIDI channel block for instrument instance
        * owning the calling opcode; this may be NULL in case of a
        * score note.
        */
        MCHNBLK *csoundGetMIDIParams(void *p);

        /**
        * Query ftable *argp at init time.
        * An init error is raised in case of invalid argument.
        */
        FUNC  *csoundQueryFTableInit(void *p, MYFLT *argp);

        /**
        * Query ftable *argp at perf time.
        * A perf error is raised in case of invalid argument.
        */
        FUNC  *csoundQueryFTablePerf(void *p, MYFLT *argp);

        /**
        * Query deferred size ftable *argp at init time.
        * An init error is raised in case of invalid argument.
        */
        FUNC  *csoundQueryDeferredFTableInit(void *p, MYFLT *argp);

        /**
        * Get release time for currently performing instrument
        * instance in seconds.
        */
        MYFLT csoundGetReleaseTime(void *p);

        /**
        * Set release time (in seconds) for current instrument instance.
        */
        void csoundSetReleaseTime(void *p, MYFLT xtim);

        /**
        * Returns non-zero if the current instrument is releasing.
        */
        int csoundGetReleaseFlag(void *p);

        /**
        * Returns non-zero if beat mode is enabled.
        */
        int csoundGetBeatModeFlag(void *p);

        /**
        * Returns the turnoff time of current instrument instance
        * in seconds or beats, depending on beat mode. Return value
        * is negative if the note has indefinite length.
        * The current time in seconds can be calculated as
        *   p->h.globals->global_onedkr * p->h.globals->global_kcounter
        * Note: beat mode is not very well tested and may not work
        * correctly in some cases.
        */
        MYFLT csoundGetOffTime(void *p);

        /**
        * Set the turnoff time of current instrument instance
        * in seconds or beats, depending on beat mode.
        */
        void csoundSetOffTime(void *p, MYFLT offtim);

        /**
        * Get pointer to the array of p-fields for the current
        * instrument instance. Returns the number of p-fields.
        * Note: pflds[1] is p1
        */
        int csoundGetPFields(void *p, MYFLT **pflds);

        /**
        * Skip rest of init opcodes for the current instrument instance.
        */
        void csoundSkipInitOpcodes(void *p);

        /**
        * Skip rest of perf opcodes for the current instrument instance.
        */
        void csoundSkipPerfOpcodes(void *p);

        /**
        * Turn off currently performing instrument instance (with release).
        * Should be used at performance time only.
        */
        void csoundTurnOffCurrentNote(void *p);

        /**
        * Allocate memory for opcode (cleared to zero).
        * Should be called at i-time only.
        */
        void csoundAuxAlloc(void *p, AUXCH *auxchp, size_t nbytes);

        /**
        * Open file (automatically closed when note is deactivated).
        * The allowed values for 'mode' are:
        *   0: open for input (read from current directory, SSDIR, or SFDIR)
        *   1: open for output (write to current directory)
        *   2: open for output (write to SFDIR)
        *   3: open for output (write to SFDIR, or current directory)
        * Returns file descriptor.
        * Should be called at init time only.
        */
        int csoundOpenFile(void *p, FDCH *fdchp, const char *name, int mode);

        /**
        * Close file opened with csoundOpenFile().
        * Should be called at init time only.
        */
        void csoundCloseFile(void *p, FDCH *fdchp);

        /**
        * Load an entire file into memory.
        * If not fullpath, look in current directory,
        * then SADIR (if defined).
        */
        MEMFIL *csoundLoadFile(void *p, char *filnam);

        /**
        * Returns a random number between 0 and 1.
        */
        MYFLT csoundRandomValueU(void *p);

        /**
        * Returns a random number between -1 and 1.
        */
        MYFLT csoundRandomValueB(void *p);

        /**
        * Set the random seed of the Csound instance owning the current
        * instrument to the specified unsigned 32 bit value.
        */
        void csoundSetRandomSeed(void *p, unsigned int seedValue);

        /**
        *
        * TIME FUNCTIONS
        *
        */

        /**
        * Initialise a timer structure.
        */

        void timers_struct_init(RTCLOCK*);

        /**
        * Returns the elapsed real time (in seconds) since the
        * specified timer structure was initialised.
        */
        double timers_get_real_time(RTCLOCK*);

        /**
        * Returns the elapsed CPU time (in seconds) since the
        * specified timer structure was initialised.
        */
        double timers_get_CPU_time(RTCLOCK*);

        /**
        * Returns a 32-bit unsigned integer to be used as
        * seed from current time.
        */
        unsigned long timers_random_seed(void);

        /**
        *
        * MEMORY MANAGEMENT
        *
        */

        /**
        * Initialise a heap for Csound instance 'csound' with the specified
        * flags; 'flags' is a bitwise OR of any of the following values:
        *   MEMALLOC_ERROR_RETURN
        *     alloc/free routines should return in case of a failure (with
        *     a NULL return value if applicable), rather than calling the
        *     csoundError() callback of 'csound' with an error message.
        *     If not set, alloc functions never return NULL (with the
        *     exception of realloc being called with a zero size).
        *   MEMALLOC_DISPOSABLE
        *     enable advanced management of allocated memory:
        *       * csoundHeapDestroy() can free all the allocated memory
        *       * improved error checking
        *     however, this makes the function calls slower, and increases
        *     memory usage by 32 to 48 bytes per allocated memory block
        * Return value is a handle for the newly created heap, which may
        * be NULL if the MEMALLOC_ERROR_RETURN flag is set and an error
        * occured.
        */
        void *csoundHeapCreate(void *csound, int flags);

        /**
        * Set error code with which CSOUNDAPI.Error is called as
        *   CSOUNDAPI.Error(csound, errCode, errMsg, handle)
        * for heap 'handle'.
        * The default error code is CSOUND_MEMORY.
        */
        void csoundHeapSetErrorCode(void *handle, int errCode);

        /**
        * Allocate a memory block for the heap identified by 'handle'.
        * 'nbytes' is the number of bytes to allocate.
        * Return value is a pointer to the new memory area.
        */
        void *csoundMalloc(void *handle, size_t nbytes);

        /**
        * Allocate a memory block for the heap identified by 'handle',
        * and clear it to zero.
        * 'nmemb' and 'size' are the same as the parameters for calloc().
        * Return value is a pointer to the new memory area.
        */
        void *csoundCalloc(void *handle, size_t nmemb, size_t size);

        /**
        * Free a memory block in heap identified by 'handle'.
        * 'ptr' is a pointer returned by a previous call to
        * csoundMalloc(), csoundCalloc(), or csoundRealloc()
        * with the same handle.
        */
        void csoundFree(void *handle, void *ptr);

        /**
        * Re-allocate a memory block for the heap identified by 'handle'.
        * 'ptr' and 'nbytes' are the same as the parameters for realloc()
        * Return value is a pointer to the new memory area.
        */
        void *csoundRealloc(void *handle, void *ptr, size_t nbytes);

        /**
        * Free all memory used by the specified heap.
        * Has no effect if called with handle = NULL.
        * WARNING: the expected input parameter is a pointer
        * to the variable storing the heap handle, and not
        * the handle itself; this variable is set to NULL
        * on return.
        */
        void csoundHeapDestroy(void **handle);

        /**
        *
        * NATIVE LANGUAGE SUPPORT
        *
        */

#include "n_getstr.h"   /* for language codes */

        /**
        * Set language to 'lang_code' (lang_code can be for example
        * CSLANGUAGE_ENGLISH_UK or CSLANGUAGE_FRENCH or many others,
        * see n_getstr.h for the list of languages). This affects all
        * Csound instances running in the address space of the current
        * process. The special language code CSLANGUAGE_DEFAULT can be
        * used to disable translation of messages and free all memory
        * allocated by a previous call to csoundSetLanguage().
        * csoundSetLanguage() loads all files for the selected language
        * from the directory specified by the CSSTRNGS environment
        * variable.
        */
        void csoundSetLanguage(cslanguage_t lang_code);

        /**
        * Translate string 's' to the current language, and return
        * pointer to the translated message. This may be the same as
        * 's' if language was set to CSLANGUAGE_DEFAULT.
        */
        char *csoundLocalizeString(const char *s);

#endif          /* !CSOUND_PLUGIN_LIBRARY */

#ifdef __cplusplus
};
#endif /* __cplusplus */

#endif /* CSOUND_H */

