
/* ALSA real-time audio routines -- written by Istvan Varga, Nov 2002 */

#include <Csound/csdl.h>
#include <alsa/asoundlib.h>
#include <unistd.h>

#undef ENGINE
#define ENGINE  ((CSOUNDAPI*) csound)
#undef Str
#define Str(x)  (((CSOUNDAPI*) csound)->LocalizeString(x))
#undef exit

typedef struct devparams_ {
        char                *device;            /* device name              */
        snd_pcm_format_t    format;             /* sample format            */
        int                 sample_size;        /* sample size in bytes     */
        int                 period_size;        /* sample buffer size       */
        int                 srate;              /* sample rate in Hz        */
        int                 nchns;              /* number of channels       */
        snd_pcm_sframes_t   buffer_smps;        /* buffer length in samples */
        snd_pcm_sframes_t   period_smps;        /* period time in samples   */
        snd_pcm_t           *handle;            /* handle                   */
        snd_pcm_hw_params_t *hw_params;         /* hardware parameters      */
        snd_pcm_sw_params_t *sw_params;         /* software parameters      */
        snd_pcm_access_t    access;             /* access method            */
} DEVPARAMS;

typedef struct {
    CSOUNDAPI   *csound;
    DEVPARAMS   in_dev;
} WAVE_IN_DATA;

typedef struct {
    CSOUNDAPI   *csound;
    DEVPARAMS   out_dev;
} WAVE_OUT_DATA;

/* select sample format */

static void set_format(void *csound, DEVPARAMS *dev, AUDIOPARMS *devParms)
{
    short   endian_test = 0x1234;

    if (*((unsigned char*) (&endian_test)) == (unsigned char) 0x34) {
      /* little-endian */
      switch (devParms->sampleFormat) {
        case AE_CHAR:   dev->format = SND_PCM_FORMAT_S8;              break;
        case AE_ALAW:   dev->format = SND_PCM_FORMAT_A_LAW;           break;
        case AE_ULAW:   dev->format = SND_PCM_FORMAT_MU_LAW;          break;
        case AE_SHORT:  dev->format = SND_PCM_FORMAT_S16_LE;          break;
        case AE_LONG:   dev->format = SND_PCM_FORMAT_S32_LE;          break;
        case AE_FLOAT:  dev->format = SND_PCM_FORMAT_FLOAT_LE;        break;
        case AE_UNCH:   dev->format = SND_PCM_FORMAT_U8;              break;
        case AE_24INT:  dev->format = SND_PCM_FORMAT_S24_LE;          break;
        case AE_DOUBLE: dev->format = SND_PCM_FORMAT_FLOAT64_LE;      break;
        default:        ENGINE->Die(csound, Str("unknown sample format"));
      }
    }
    else {
      /* big-endian */
      switch (devParms->sampleFormat) {
        case AE_CHAR:   dev->format = SND_PCM_FORMAT_S8;              break;
        case AE_ALAW:   dev->format = SND_PCM_FORMAT_A_LAW;           break;
        case AE_ULAW:   dev->format = SND_PCM_FORMAT_MU_LAW;          break;
        case AE_SHORT:  dev->format = SND_PCM_FORMAT_S16_BE;          break;
        case AE_LONG:   dev->format = SND_PCM_FORMAT_S32_BE;          break;
        case AE_FLOAT:  dev->format = SND_PCM_FORMAT_FLOAT_BE;        break;
        case AE_UNCH:   dev->format = SND_PCM_FORMAT_U8;              break;
        case AE_24INT:  dev->format = SND_PCM_FORMAT_S24_BE;          break;
        case AE_DOUBLE: dev->format = SND_PCM_FORMAT_FLOAT64_BE;      break;
        default:        ENGINE->Die(csound, Str("unknown sample format"));
      }
    }
    /* calculate sample size */
    dev->sample_size = devParms->sampleSize;
}

/* set up audio device */

static void set_device_params(void *csound,
                              DEVPARAMS *dev, AUDIOPARMS *devParms, int output)
{
    int     err, dir, n;
    char    msg[1024], *device_name = devParms->deviceName;

    /* open the device ... */
    if (!device_name)
      device_name = "default";
    if (output) {
      /* ... for playback */
      err = snd_pcm_open(&(dev->handle), device_name,
                         SND_PCM_STREAM_PLAYBACK, 0);
      if (err < 0) {
        sprintf(msg, Str("cannot open device %s for audio output: %s"),
                     device_name, snd_strerror(err));
        ENGINE->Die(csound, msg);
        return;
      }
    }
    else {
      /* ... for capture */
      err = snd_pcm_open(&(dev->handle), device_name,
                         SND_PCM_STREAM_CAPTURE, 0);
      if (err < 0) {
        sprintf(msg, Str("cannot open device %s for audio input: %s"),
                     device_name, snd_strerror(err));
        ENGINE->Die(csound, msg);
        return;
      }
    }
    /* allocate hardware and software parameters */
    snd_pcm_hw_params_alloca(&(dev->hw_params));
    snd_pcm_sw_params_alloca(&(dev->sw_params));
    if (snd_pcm_hw_params_any(dev->handle, dev->hw_params) < 0) {
      ENGINE->Die(csound, Str("no real-time audio configurations found"));
      return;
    }
    /* now set the various hardware parameters: */
    /* access method, */
    dev->access = SND_PCM_ACCESS_RW_INTERLEAVED;
    if (snd_pcm_hw_params_set_access(dev->handle, dev->hw_params, dev->access)
        < 0) {
      ENGINE->Die(csound, Str("error setting access type for soundcard"));
      return;
    }
    /* sample format, */
    set_format(csound, dev, devParms);
    if (snd_pcm_hw_params_set_format(dev->handle, dev->hw_params, dev->format)
        < 0) {
      ENGINE->Die(csound,
                  Str("unable to set requested sample format on soundcard"));
      return;
    }
    /* number of channels, */
    dev->nchns = devParms->nChannels;
    if (snd_pcm_hw_params_set_channels(dev->handle, dev->hw_params,
                                       devParms->nChannels) < 0) {
      ENGINE->Die(csound, Str("unable to set mode (mono/stereo) on soundcard"));
      return;
    }
    dev->sample_size *= devParms->nChannels;    /* correct sample size */
    /* sample rate, */
    dev->srate =
      (int) snd_pcm_hw_params_set_rate_near(dev->handle, dev->hw_params,
                                            (int) (devParms->sampleRate + 0.5f),
                                            0);
    if (dev->srate < 0) {
      ENGINE->Die(csound, Str("unable to set sample rate on soundcard"));
      return;
    }
    if ((ENGINE->GetMessageLevel(csound) & 4) &&
        dev->srate != (int) (devParms->sampleRate + 0.5f)) {
      sprintf(msg, Str("Sample rate set to %d (instead of %d)\n"),
                   dev->srate, (int) (devParms->sampleRate + 0.5f));
      ENGINE->Message(csound, msg);
    }
    /* buffer size, */
    n = devParms->hwBufSize;
    dev->buffer_smps = (snd_pcm_sframes_t)
      snd_pcm_hw_params_set_buffer_size_near(dev->handle, dev->hw_params,
                                             (snd_pcm_uframes_t) n);
    if (dev->buffer_smps <= (snd_pcm_sframes_t) 0) {
      ENGINE->Die(csound,
                  Str("failed while trying to set soundcard DMA buffer size"));
      return;
    }
    if ((int) dev->buffer_smps != n && (ENGINE->GetMessageLevel(csound) & 4)) {
      sprintf(msg, "Soundcard DMA buffer size set to %d samples "
                   "(instead of %d)\n",
              (int) dev->buffer_smps, n);
      ENGINE->Message(csound, msg);
    }
    /* and period size */
    n = devParms->swBufSize;
    dev->period_smps = (snd_pcm_sframes_t)
      snd_pcm_hw_params_set_period_size_near(dev->handle, dev->hw_params,
                                             (snd_pcm_uframes_t) n, &dir);
    if (dev->period_smps <= 0) {
      ENGINE->Die(csound,
                  Str("error setting period time for real-time audio"));
      return;
    }
    if ((int) dev->period_smps != n && (ENGINE->GetMessageLevel(csound) & 4)) {
      sprintf(msg, "Real-time audio period size set to %d samples "
                   "(instead of %d)\n",
              (int) dev->period_smps, n);
      ENGINE->Message(csound, msg);
    }
    dev->period_size = (int) dev->period_smps * dev->sample_size
                       * devParms->nChannels;
    /* set up device according to the above parameters */
    if (snd_pcm_hw_params(dev->handle, dev->hw_params) < 0) {
      ENGINE->Die(csound,
                  Str("error setting hardware parameters for real-time audio"));
      return;
    }
    /* now set software parameters */
    n = (output ? dev->buffer_smps : 1);
    if (snd_pcm_sw_params_current(dev->handle, dev->sw_params) < 0
        || snd_pcm_sw_params_set_start_threshold(dev->handle, dev->sw_params,
                                                 (snd_pcm_uframes_t) n) < 0
        || snd_pcm_sw_params_set_avail_min(dev->handle, dev->sw_params,
                                           dev->period_smps) < 0
        || snd_pcm_sw_params_set_xfer_align(dev->handle, dev->sw_params, 1) < 0
        || snd_pcm_sw_params(dev->handle, dev->sw_params) < 0) {
      ENGINE->Die(csound,
                  Str("error setting software parameters for real-time audio"));
    }
}

/* open for audio input */

static void *recopen(void *csound, AUDIOPARMS *devParms)
{
    WAVE_IN_DATA    *audioInData;

    /* allocate instance data */
    audioInData =
      (WAVE_IN_DATA*) calloc((size_t) 1, (size_t) sizeof(WAVE_IN_DATA));
    if (audioInData == NULL) {
      ENGINE->Die(csound, "not enough memory");
      return NULL;
    }
    audioInData->csound = (CSOUNDAPI*) csound;
    set_device_params(csound, &(audioInData->in_dev), devParms, 0);

    return (void*) audioInData;
}

/* open for audio output */

static void *playopen(void *csound, AUDIOPARMS *devParms)
{
    WAVE_OUT_DATA   *audioOutData;

    /* allocate instance data */
    audioOutData =
      (WAVE_OUT_DATA*) calloc((size_t) 1, (size_t) sizeof(WAVE_OUT_DATA));
    if (audioOutData == NULL) {
      ENGINE->Die(csound, "not enough memory");
      return NULL;
    }
    audioOutData->csound = (CSOUNDAPI*) csound;
    set_device_params(csound, &(audioOutData->out_dev), devParms, 1);

    return (void*) audioOutData;
}

/* get samples from ADC */

static int rtrecord(WAVE_IN_DATA *audioInData, char *inbuf, int nbytes)
{
    int     n = nbytes / audioInData->in_dev.sample_size;   /* nr of samples */
    int     m = 0, err;

    while (n) {
      err = (int) snd_pcm_readi(audioInData->in_dev.handle, (void*) inbuf,
                                (snd_pcm_uframes_t) n);
      if (err >= 0) {
        n -= err; m += err; continue;
      }
      else {
        CSOUNDAPI   *csound = audioInData->csound;
        /* handle I/O errors */
        if (err == -EPIPE) {
          /* buffer underrun */
          if (csound->GetMessageLevel((void*) csound) & 4)
            csound->Message((void*) csound,
              Str("Buffer underrun in real-time audio input\n")); /* complain */
          if (snd_pcm_prepare(audioInData->in_dev.handle) >= 0) continue;
        }
        else if (err == -ESTRPIPE) {
          /* suspend */
          if (csound->GetMessageLevel((void*) csound) & 4)
            csound->Message((void*) csound,
              Str("Real-time audio input suspended\n"));
          while (snd_pcm_resume(audioInData->in_dev.handle) == -EAGAIN)
            sleep(1);
          if (snd_pcm_prepare(audioInData->in_dev.handle) >= 0) continue;
        }
        /* could not recover from error */
        csound->Die((void*) csound,
                    Str("Error reading data from audio input device"));
        return 0;
      }
    }

    return (m * audioInData->in_dev.sample_size);
}

/* put samples to DAC */

static void rtplay(WAVE_OUT_DATA *audioOutData, char *outbuf, int nbytes)
{
    int     n = nbytes / audioOutData->out_dev.sample_size; /* nr of samples */
    int     err;

    while (n) {
      err = (int) snd_pcm_writei(audioOutData->out_dev.handle, (void*) outbuf,
                                 (snd_pcm_uframes_t) n);
      if (err >= 0) {
        n -= err; continue;
      }
      else {
        CSOUNDAPI *csound = audioOutData->csound;
        /* handle I/O errors */
        if (err == -EPIPE) {
          /* buffer underrun */
          if (csound->GetMessageLevel((void*) csound) & 4)
            csound->Message((void*) csound,
              Str("Buffer underrun in real-time audio output\n"));/* complain */
          if (snd_pcm_prepare(audioOutData->out_dev.handle) >= 0)
            continue;
        }
        else if (err == -ESTRPIPE) {
          if (csound->GetMessageLevel((void*) csound) & 4)  /* suspend */
            csound->Message((void*) csound,
                            Str("Real-time audio output suspended\n"));
          while (snd_pcm_resume(audioOutData->out_dev.handle) == -EAGAIN)
            sleep(1);
          if (snd_pcm_prepare(audioOutData->out_dev.handle) >= 0) continue;
        }
        /* could not recover from error */
        csound->Die((void*) csound,
                    Str("Error writing data to audio output device"));
        return;
      }
    }
}

/* close the I/O device entirely  */
/* called only when both complete */

static void rtclose(WAVE_IN_DATA *audioInData, WAVE_OUT_DATA *audioOutData)
{
    if (audioInData != NULL) {
      snd_pcm_close(audioInData->in_dev.handle);
      free(audioInData);
    }
    if (audioOutData != NULL) {
      snd_pcm_close(audioOutData->out_dev.handle);
      free(audioOutData);
    }
}

long opcode_count(void *csound)
{
    csound = csound;
    return 0L;
}

void *opcode_init(void *csound)
{
    fprintf(stderr, "ALSA 0.9.x plugin by Istvan Varga\n");

    ENGINE->SetPlayopenCallback(csound, playopen);
    ENGINE->SetRtplayCallback(csound, (void (*)(void*, char*, int)) rtplay);
    ENGINE->SetRecopenCallback(csound, recopen);
    ENGINE->SetRtrecordCallback(csound, (int (*)(void*, char*, int)) rtrecord);
    ENGINE->SetRtcloseCallback(csound, (void (*)(void*, void*)) rtclose);

    return NULL;
}

